/*
 * ext.c
 *
 *  Created on: 25.06.2019
 *      Author: AndreR
 */

#include "ext.h"
#include "main/robot.h"
#include "util/init_hal.h"
#include "util/log.h"
#include "util/sys_time.h"
#include "constants.h"
#include "errors.h"
#include "commands.h"
#include <string.h>

static void processReceivedData(const uint8_t* pData, uint32_t dataLength);

Ext ext;

void USART6_IRQHandler()
{
	CH_IRQ_PROLOGUE();
	UARTFifoIRQ(&ext.uart);
	CH_IRQ_EPILOGUE();
}

// called at IRQ level!
static void usart6RxIrqHook()
{
	chSysLockFromISR();
	chBSemSignalI(&ext.dataAvailable);
	chSysUnlockFromISR();
}

void ExtInit(uint32_t baudrate)
{
	ext.baudrate = baudrate;

	chBSemObjectInit(&ext.dataAvailable, 1);

	FifoLinInit(&ext.txFifo, ext.txData, EXT_TX_FIFO_SIZE);

	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 7;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_2MHZ;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOG, GPIO_PIN_9 | GPIO_PIN_14, &gpioInit);
	RCC->APB2ENR |= RCC_APB2ENR_USART6EN;
	__DSB();

	UARTFifoInit(&ext.uart, USART6, baudrate, systemClockInfo.APB2PeriphClk, 0);
	ext.uart.rxIrqHook = usart6RxIrqHook;

	NVICEnableIRQ(USART6_IRQn, IRQL_EXT_COMM);
}

void ExtTask(void* params)
{
	(void)params;

	chRegSetThreadName("EXT");

	int16_t result;
	uint32_t last1sTime = 0;

	while(1)
	{
		chBSemWaitTimeout(&ext.dataAvailable, MS2ST(10));

		uint8_t* pTxData;
		uint32_t txSize;

		// send all packets in transmit buffer
		while(FifoLinGet(&ext.txFifo, &pTxData, &txSize) == 0)
		{
			ext.liveStats.txPayload += txSize;

			uint32_t encodedSize;
			if(COBSEncode(pTxData, txSize, ext.txProcessingBuf, EXT_MAX_PACKET_SIZE, &encodedSize) == 0)
			{
				ext.liveStats.txWire += encodedSize+1;

				ext.txProcessingBuf[encodedSize] = 0;

				result = UARTFifoWrite(&ext.uart, ext.txProcessingBuf, encodedSize+1, MS2ST(100));
				if(result)
					LogErrorC("UARTFifo write error", result);
			}
			else
			{
				LogWarn("COBS Error");
			}

			FifoLinDelete(&ext.txFifo);
		}

		// process received data
		uint16_t rxStatus;
		uint16_t rxSize = sizeof(ext.rxReceiveBuf);
		while(UARTFifoRead(&ext.uart, ext.rxReceiveBuf, &rxSize, &rxStatus, 0) == 0)
		{
			if(rxStatus & UART_FIFO_STATUS_ERR_MASK)
			{
				LogWarnC("Errornous packet received", rxStatus);
				continue;
			}

			ext.liveStats.rxWire += rxSize;

			for(uint16_t i = 0; i < rxSize; i++)
			{
				uint8_t c = ext.rxReceiveBuf[i];

				if(c == 0)
				{
					// packet complete
					uint32_t decodedSize;
					if(COBSDecode(ext.rxProcessingBuf, ext.rxProcBufUsed, ext.rxDecodeBuf, EXT_MAX_PACKET_SIZE, &decodedSize) == 0)
					{
						ext.liveStats.rxPayload += decodedSize;

						processReceivedData(ext.rxDecodeBuf, decodedSize);
					}
					else
					{
						LogWarn("COBS decode error");
					}

					ext.rxProcBufUsed = 0;
				}
				else
				{
					if(ext.rxProcBufUsed == sizeof(ext.rxProcessingBuf))
					{
						// we filled the whole buffer without delimiter => start again
						ext.rxProcBufUsed = 0;
						LogWarn("rx proc buf overrun");
					}

					ext.rxProcessingBuf[ext.rxProcBufUsed++] = c;
				}
			}
		}

		if(chVTTimeElapsedSinceX(last1sTime) > S2ST(1))
		{
			last1sTime = chVTGetSystemTimeX();

			if(ext.stats.rxWire > 0)
				ext.installed = 1;
			else
				ext.installed = 0;

			memcpy(&ext.stats, &ext.liveStats, sizeof(EXTStats));
			memset(&ext.liveStats, 0, sizeof(EXTStats));
			ext.stats.rxUsage = ((float)ext.stats.rxWire)/(ext.baudrate*0.1f);
			ext.stats.txUsage = ((float)ext.stats.txWire)/(ext.baudrate*0.1f);
		}
	}
}


int16_t ExtSendPacket(PacketHeader* pHeader, const void* _pData, uint16_t dataLength)
{
	if(!ext.installed)
		return ERROR_INVALID_HANDLE;

	uint8_t* pData = (uint8_t*)_pData;

	uint8_t* pDst;
	int16_t result;

	result = FifoLinReserve(&ext.txFifo, dataLength + PACKET_HEADER_SIZE, &pDst);
	if(result)
		return result;

	memcpy(pDst, pHeader->_data, PACKET_HEADER_SIZE);
	memcpy(pDst + PACKET_HEADER_SIZE, pData, dataLength);

	result = FifoLinCommit(&ext.txFifo, dataLength);
	if(result)
		return result;

	chBSemSignal(&ext.dataAvailable);

	return 0;
}

static void processReceivedData(const uint8_t* pData, uint32_t dataLength)
{
	if(dataLength < 2)
	{
		LogErrorC("Packet too short", dataLength);
		return;
	}

	PacketHeader* pHeader = (PacketHeader*)pData;
	pData += sizeof(PacketHeader);
	dataLength -= sizeof(PacketHeader);

	if(pHeader->section == SECTION_SYSTEM && pHeader->cmd == CMD_SYSTEM_MATCH_CTRL)
	{
		if(dataLength < sizeof(SystemMatchCtrl)-SYSTEM_MATCH_CTRL_USER_DATA_SIZE)
		{
			LogWarnC("Invalid match ctrl size from EXT", dataLength);
		}
		else
		{
			memcpy(&robot.ext.matchCtrl, pData, dataLength);
			robot.ext.lastMatchCtrlTime = SysTimeUSec();
		}
	}
}
