/*
 * ir.c
 *
 *  Created on: 12.01.2019
 *      Author: AndreR
 */

#include "ir.h"

#include "util/init_hal.h"
#include "hal/port_ex.h"
#include "hal/kicker.h"
#include "util/log.h"
#include "util/console.h"
#include "constants.h"
#include <string.h>


IR ir;

void UART8_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	UARTFifoIRQ(&ir.uart);

	CH_IRQ_EPILOGUE();
}

static void startBootloader(uint8_t target)
{
	PortExBootSet(target, 1);
	PortExResetPulse(target);
	PortExBootSet(target, 0);
}

void IrInit()
{
	chMtxObjectInit(&ir.mtxUart);

	// init port
	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_2MHZ;
	gpioInit.alternate = 8;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOE, GPIO_PIN_0 | GPIO_PIN_1, &gpioInit);

	// UART8
	RCC->APB1LENR |= RCC_APB1LENR_UART8EN;
	__DSB();

	UARTFifoInit(&ir.uart, UART8, 2000000, systemClockInfo.APB1PeriphClk, 1);

	NVICEnableIRQ(UART8_IRQn, IRQL_UART8);

	STBootloaderInit(&ir.bootloader, &ir.uart, &startBootloader, PORT_EX_TARGET_IR);
}

void IrFlashProgram(uint8_t force)
{
	chMtxLock(&ir.mtxUart);

	uint32_t originalBaudrate = ir.uart.baudrate;
	UARTFifoSetBaudrate(&ir.uart, 115200);
	ir.flashFuncResult = STBootloaderFlash(&ir.bootloader ,(uint8_t*)0x081A0000, 32*1024, 0x08000000, force, &ir.flashResult);
	UARTFifoSetBaudrate(&ir.uart, originalBaudrate);

	chMtxUnlock(&ir.mtxUart);

	ConsolePrint("IR  %hu    %hu    0x%04hX  %-6u%4u     %ums\r\n",
			(uint16_t)ir.flashResult.updateRequired, (uint16_t)ir.flashResult.updated,
			ir.flashFuncResult, ir.flashResult.programSize, ir.flashResult.numBlocks,
			ir.flashResult.timeUs/1000);
}

void IrTask(void* params)
{
	(void)params;

	chRegSetThreadName("IR");

	chThdSleepMilliseconds(100);

	IrFlashProgram(0);

	IrExchangeMISO miso;
	uint16_t rxDatasize;
	uint16_t fifoStatus;

	while(1)
	{
		chMtxLock(&ir.mtxUart);
		msg_t waitResult = UARTFifoReadWait(&ir.uart, MS2ST(20));
		chMtxUnlock(&ir.mtxUart);
		if(waitResult != MSG_OK)
		{
			LogWarn("rx timeout");
			continue;
		}

		int16_t result;
		while(1)
		{
			rxDatasize = sizeof(IrExchangeMISO);
			chMtxLock(&ir.mtxUart);
			result = UARTFifoRead(&ir.uart, (uint8_t*)&miso, &rxDatasize, &fifoStatus, 0);
			chMtxUnlock(&ir.mtxUart);
			if(result == UART_FIFO_ERROR_NO_DATA)
				break;
			if(result)
			{
				LogWarnC("read error", result);
				continue;
			}
			if(rxDatasize < sizeof(IrExchangeMISO))
			{
				LogWarnC("Short read", rxDatasize);
				continue;
			}

			uint8_t* pIrData = (uint8_t*)miso.channels;
			uint8_t xor = 0;
			for(uint16_t i = 0; i < sizeof(miso.channels)+sizeof(miso.config)+sizeof(miso.xorChecksum); i++)
				xor ^= pIrData[i];

			if(xor != 0)
			{
				LogWarnC("Checksum error", xor);
				continue;
			}

			uint8_t channel = miso.config;
			if(channel > 5)
			{
				LogWarnC("Invalid IR channel", channel);
				continue;
			}

			ir.vBarrier = ((float)miso.channels[0]) * 1.0f/8.0f * 8.056640625e-4f;
			for(uint8_t i = 1; i < 7; i++)
			{
				ir.vLateral[channel][i-1] = ((float)miso.channels[i]) * 1.0f/8.0f * 8.056640625e-4f;
			}

			KickerIrUpdate(ir.vBarrier, 0.0f);

			ir.cpuLoad = ((float)(miso.idleTicksNoLoad - miso.idleTicksActive))/((float)miso.idleTicksNoLoad)*100.0f;
		}
	}
}
