/*
 * motors.c
 *
 *  Created on: 05.01.2019
 *      Author: AndreR
 */

#include "motors.h"
#include "motors_com.h"
#include "util/init_hal.h"
#include "util/console.h"
#include "util/log.h"
#include "constants.h"
#include "main/ctrl.h"
#include "struct_ids.h"
#include "power.h"

// hall speed conversion:
// hallSpeed = rads * 1/(2*float(%pi))*pp*1.0/48000000*2^32*1/2^7;
// => rad per sec = hallSpeed * 0.1872535141461965 (drive motors, 48 commutations per revolution)
// => rad per sec = hallSpeed * 1.498028113169572 (dribbler motor, 6 commutations per revolution)
#define HALL_SPEED_TO_RAD_PER_SEC_DRIVE (0.1872535141461965f)
#define RAD_PER_SEC_DRIVE_TO_HALL_SPEED (5.34035371544087f)
#define HALL_SPEED_TO_RAD_PER_SEC_DRIBBLER (1.498028113169572f)
#define RAD_PER_SEC_DRIBBLER_TO_HALL_SPEED (0.6675442144301087f)

// encoder speed conversion:
// encDelta = rads * 1/(2*float(%pi)) * 36864 * 1/1000;
// => encDelta = 5.86708782213963 * rad per sec
// => rad per sec = encDelta * 0.1704423097650712
//#define ENCODER_DELTA_TO_RAD_PER_SEC (0.1704423097650712f)
//#define RAD_PER_SEC_TO_ENCODER_DELTA (5.86708782213963)

static inline void driveCtrl();
static inline void dribblerCtrl();

// data holders
Motors motors = {
	.config = {
		.Kp = 0.05f,
		.Ki = 2.0f,
		.Kd = 0.00f,
		.Ktp = 0.0f,
		.Ktd = 0.0f,
		.K = 35.0f,
		.T = 0.1f,
	},
};

static const ConfigFileDesc configFileDescMotors =
	{ SID_CFG_MOTORS, 1, "motors", 7, (ElementDesc[]) {
		{ FLOAT, "pid_p", "", "PID/p" },
		{ FLOAT, "pid_i", "", "PID/i" },
		{ FLOAT, "pid_d", "", "PID/d" },
		{ FLOAT, "pid_tp", "s", "PID/tp" },
		{ FLOAT, "pid_td", "s", "PID/td" },
		{ FLOAT, "model_K", "", "model/K" },
		{ FLOAT, "model_T", "s", "model/T" },
	} };

static void configUpdateCallback(uint16_t cfgId)
{
	if(cfgId == SID_CFG_MOTORS)
	{
		for(uint8_t i = 0; i < 4; i++)
		{
			PIDSetParams(&motors.drive[i].pid, motors.config.Kp, motors.config.Ki, motors.config.Kd);
			PIDSetPFilterParam(&motors.drive[i].pid, motors.config.Ktp);
			PIDSetDFilterParam(&motors.drive[i].pid, motors.config.Ktd);
		}
	}
}

// ####################
// Public functions
// ####################
void MotorsInit()
{
	motors.driveMode = MOT_CTRL_OFF;
	motors.dribbler.mode = DRIBBLER_MODE_OFF;

	PIDParams params;
	params.kp = 0.10f; // 0.15f
	params.ki = 10.0f; //15.0f
	params.kd = 0.0f;
	params.outputMin = -10.0f;	// [V]
	params.outputMax = 10.0f;	// [V]
	params.dFilterTimeConstant = 0.0f;
	params.pFilterTimeConstant = 0.0f;

	for(uint8_t i = 0; i < 4; i++)
	{
		PIDInit(&motors.drive[i].pid, &params, 0.001f, 1);
	}

	params.kp = 0.002f;
	params.ki = 0.05f;
	params.kd = 0.0f;
	params.outputMin = 0.0f;	// [V]
	PIDInit(&motors.dribbler.pid, &params, 0.01f, 1);

	LagElementPT1Init(&motors.dribbler.tempEstimator, 6.0f, 350.0f, 0.001f);

	motors.pConfigFileMotors = ConfigOpenOrCreate(&configFileDescMotors, &motors.config, sizeof(MotorsConfig), &configUpdateCallback, 0);

	ConfigNotifyUpdate(motors.pConfigFileMotors);
}

void MotorsTask(void* params)
{
	(void)params;

	chRegSetThreadName("Motors");

	systime_t prev = chVTGetSystemTimeX();
	systime_t next = prev+US2ST(10000);

	while(1)
	{
		MotorsComRunStatemachine();

		// get new data from motors
		MotorsComReceiveData();

		// run control
		driveCtrl();
		dribblerCtrl();

		// send new data to motors
		MotorsComStartSendData();
		MotorsComWaitSendDataComplete();

		chThdSleepUntilWindowed(prev, next);
		prev = next;
		next += US2ST(1000);
	}
}

void MotorsDriveOff()
{
	motors.driveMode = MOT_CTRL_OFF;
}

void MotorsDriveSetVoltage(uint8_t motor, float vol)
{
	motors.driveMode = MOT_CTRL_VOLTAGE;
	motors.drive[motor].setVoltage = vol;
}

void MotorsDriveSetVelocity(uint8_t motor, float vel)
{
	motors.driveMode = MOT_CTRL_VELOCITY;
	motors.drive[motor].setVelocity = vel;
}

void MotorsDribblerOff()
{
	motors.dribbler.mode = DRIBBLER_MODE_OFF;
}

void MotorsDribblerSetVoltage(float vol)
{
	vol *= -1.0f;

	if(vol > 12.0f)
		vol = 12.0f;

	if(vol < -12.0f)
		vol = -12.0f;

	motors.dribbler.mode = DRIBBLER_MODE_VOLTAGE;
	motors.dribbler.setVoltage = vol;
}

void MotorsDribblerSetVelocity(float velRPM)
{
	velRPM *= -1.0f;

	if(velRPM > 20000)
		velRPM = 20000;

	if(velRPM < -20000)
		velRPM = -20000;

	if(-150 < velRPM && velRPM < 150)
		velRPM = 0;

	// minimum speed is 300 RPM => ~100 Hall sensor speed updates per second
	if(-300 < velRPM && velRPM <= -150)
		velRPM = -300;

	if(150 <= velRPM && velRPM < 300)
		velRPM = 300;

	motors.dribbler.mode = DRIBBLER_MODE_SPEED;
	motors.dribbler.setVelocity = velRPM*MOTORS_DRIBBLER_BAR_RPM_TO_MOTOR_RADS;	// convert to motor rad/s
}

void MotorsTestSetAngle(uint8_t motor, float angle, float voltage)
{
	if(motor <= 3)
	{
		motors.driveMode = MOT_CTRL_ANGLE;
		motors.drive[motor].setAngle = angle;
		motors.drive[motor].setVoltage = voltage;
	}
	else
	{
		motors.dribbler.mode = MOT_CTRL_ANGLE;
		motors.dribbler.setAngle = angle;
		motors.dribbler.setVoltage = voltage;
	}
}

// ####################
// Private functions
// ####################
static inline void modelFuncMotorInv(float vel, float acc, float* pOutVol)
{
	// PT1
	*pOutVol = 1.0f/motors.config.K*vel + motors.config.T/motors.config.K*acc;
}

static inline void driveCtrl()
{
	for(uint16_t motorId = 0; motorId < 4; motorId++)
	{
		float velocity = motorsCom.motor[motorId].encoderVelocity;

		// enable/disable PID depending on mode
		if(motors.driveMode == MOT_CTRL_VELOCITY || motors.driveMode == MOT_CTRL_VEL_ACC || motors.driveMode == MOT_CTRL_VEL_TORQUE)
			PIDEnable(&motors.drive[motorId].pid, 1);
		else
			PIDEnable(&motors.drive[motorId].pid, 0);

		// always do PID control, output is not always applied depending on drive mode
		PIDSetLimits(&motors.drive[motorId].pid, -power.vBat, power.vBat);
		motors.drive[motorId].pid.setpoint = motors.drive[motorId].setVelocity;
		PIDUpdate(&motors.drive[motorId].pid, velocity);

		// set motor voltage depending on drive mode
		switch(motors.driveMode)
		{
			case MOT_CTRL_VEL_ACC:
			{
				float outVol;
				modelFuncMotorInv(motors.drive[motorId].setVelocity, motors.drive[motorId].setAcceleration, &outVol);
				MotorsComSetVoltage(motorId, motors.drive[motorId].pid.output + outVol);
			}
			break;
			case MOT_CTRL_VELOCITY:
			{
				MotorsComSetVoltage(motorId, motors.drive[motorId].pid.output);
			}
			break;
			case MOT_CTRL_VOLTAGE:
			{
				MotorsComSetVoltage(motorId, motors.drive[motorId].setVoltage);
			}
			break;
			case MOT_CTRL_TORQUE:
			{
				float setVoltage = motors.drive[motorId].setTorque * CTRL_MOTOR_R/CTRL_MOTOR_KM + CTRL_MOTOR_KE*velocity;
				MotorsComSetVoltage(motorId, setVoltage);
			}
			break;
			case MOT_CTRL_VEL_TORQUE:
			{
				float setVoltage = motors.drive[motorId].setTorque * CTRL_MOTOR_R/CTRL_MOTOR_KM + CTRL_MOTOR_KE*velocity;
				MotorsComSetVoltage(motorId, setVoltage + motors.drive[motorId].pid.output);
			}
			break;
			case MOT_CTRL_ANGLE:
			{
				MotorsComTestSetElectricalAngle(motorId, motors.drive[motorId].setAngle, motors.drive[motorId].setVoltage);
			}
			break;
			default:
			{
				MotorsComSetOff(motorId);
			}
			break;
		}
	}
}

static inline void dribblerCtrl()
{
	static uint8_t dribCnt = 0;
	static float speedVoltageOutput = 0.0f;

	float hallVelocity = motorsCom.motor[4].hallVelocity;

	// enable/disable PID depending on mode
	if(motors.dribbler.mode == DRIBBLER_MODE_SPEED)
		PIDEnable(&motors.dribbler.pid, 1);
	else
		PIDEnable(&motors.dribbler.pid, 0);

	++dribCnt;
	if(dribCnt == 10)
	{
		dribCnt = 0;

		// always do PID control, output is not always applied depending on drive mode
		PIDSetLimits(&motors.dribbler.pid, -power.vBat, power.vBat);
		motors.dribbler.pid.setpoint = motors.dribbler.setVelocity;
		PIDUpdate(&motors.dribbler.pid, hallVelocity);
	}

	motors.dribbler.temperature = LagElementPT1Process(&motors.dribbler.tempEstimator, motorsCom.motor[4].avgCurrent1ms) + 25.0f;

	if(motors.dribbler.temperature < 70.0f)
		motors.dribbler.overheated = 0;

	if(motors.dribbler.temperature > 80.0f)
		motors.dribbler.overheated = 1;

	switch(motors.dribbler.mode)
	{
		case DRIBBLER_MODE_VOLTAGE:
		{
			MotorsComSetVoltage(4, motors.dribbler.setVoltage);

			speedVoltageOutput = motors.dribbler.setVoltage;
		}
		break;
		case DRIBBLER_MODE_SPEED:
		{
			const float slope = 0.02f;
			float targetVoltage = motors.dribbler.setVelocity * 1.0f/300.0f;

			if(speedVoltageOutput < targetVoltage)
				speedVoltageOutput += slope;

			if(speedVoltageOutput > targetVoltage)
				speedVoltageOutput -= slope;

			MotorsComSetVoltage(4, speedVoltageOutput);

//			if(fabsf(motors.dribbler.setVelocity) < 1.0f && fabsf(hallVelocity) < 200.0f)
//			{
//				MotorsComSetVoltage(4, 0.0f);
//			}
//			else
//			{
//				MotorsComSetVoltage(4, motors.dribbler.pid.output);
//			}
		}
		break;
		case MOT_CTRL_ANGLE:
		{
			MotorsComTestSetElectricalAngle(4, motors.dribbler.setAngle, motors.dribbler.setVoltage);

			speedVoltageOutput = motors.dribbler.setVoltage;
		}
		break;
		default:
		{
			MotorsComSetOff(4);

			speedVoltageOutput = 0.0f;
		}
		break;
	}

	if(fabsf(motorsCom.motor[4].setVoltage) < 0.1f && fabsf(hallVelocity) < 25.0f)
		MotorsComSetOff(4);
}
