/*
 * motors_com.c
 *
 *  Created on: 05.01.2019
 *      Author: AndreR
 */

#include "motors_com.h"
#include "hal/port_ex.h"
#include "hal/buzzer.h"
#include "util/init_hal.h"
#include "util/fixed_point.h"
#include "util/console.h"
#include "util/log.h"
#include "util/angle_math.h"
#include "util/sys_time.h"
#include "constants.h"

// motor program code constants
#define MOTORS_TARGET_ADDRESS 0x08000000
#define MOTORS_SOURCE_ADDRESS 0x08180000
#define MOTORS_PROGRAM_SIZE (32*1024)

// hall speed conversion:
// hallSpeed = rads * 1/(2*float(%pi))*pp*1.0/48000000*2^32*1/2^7;
// => rad per sec = hallSpeed * 0.1872535141461965 (drive motors, 48 commutations per revolution)
// => rad per sec = hallSpeed * 1.498028113169572 (dribbler motor, 6 commutations per revolution)
#define HALL_SPEED_TO_RAD_PER_SEC_DRIVE (0.1872535141461965f)
#define RAD_PER_SEC_DRIVE_TO_HALL_SPEED (5.34035371544087f)
#define HALL_SPEED_TO_RAD_PER_SEC_DRIBBLER (1.498028113169572f)
#define RAD_PER_SEC_DRIBBLER_TO_HALL_SPEED (0.6675442144301087f)

// encoder speed conversion:
// encDelta = rads * 1/(2*float(%pi)) * 36864 * 1/1000;
// => encDelta = 5.86708782213963 * rad per sec
// => rad per sec = encDelta * 0.1704423097650712
//#define ENCODER_DELTA_TO_RAD_PER_SEC (0.1704423097650712f)
//#define RAD_PER_SEC_TO_ENCODER_DELTA (5.86708782213963)

// internal states and events
#define STATE_NORMAL		0
#define STATE_BOOTLOADER	1

#define EVENT_UPDATE_FIRMWARE		0
#define EVENT_UPDATE_FIRMWARE_FORCE	1

#define MODE_OFF 0
#define MODE_VOLTAGE 1
#define MODE_CURRENT 2
#define MODE_TEST_ANGLE 3

// data holders
MotorsCom motorsCom;

#define RECORD_SIZE 10000
static int16_t recordCurrent[RECORD_SIZE][2]  __attribute__((aligned(1024), section(".sram0_dma")));;
static uint16_t recordVoltage[RECORD_SIZE]  __attribute__((aligned(1024), section(".sram0_dma")));;
static int16_t recordSpeed[RECORD_SIZE]  __attribute__((aligned(1024), section(".sram0_dma")));;
static uint16_t recordCounter;

// forward declerations
static void startBootloader(uint8_t target);
static void flashProgram(uint16_t motorId, uint8_t force);
static void receiveMotorData(uint32_t motorId);
static void sendMotorData(uint32_t motorId);
static void flashTask(void* params);
static void parallelUpdate(uint32_t force);

// ####################
// IRQs
// ####################
void USART1_IRQHandler() // Motor 1
{
	CH_IRQ_PROLOGUE();
	UARTFifoIRQ(&motorsCom.motor[0].uart);
	CH_IRQ_EPILOGUE();
}

void UART7_IRQHandler() // Motor 2
{
	CH_IRQ_PROLOGUE();
	UARTFifoIRQ(&motorsCom.motor[1].uart);
	CH_IRQ_EPILOGUE();
}

void UART4_IRQHandler() // Motor 3
{
	CH_IRQ_PROLOGUE();
	UARTFifoIRQ(&motorsCom.motor[2].uart);
	CH_IRQ_EPILOGUE();
}

void USART3_IRQHandler() // Motor 4
{
	CH_IRQ_PROLOGUE();
	UARTFifoIRQ(&motorsCom.motor[3].uart);
	CH_IRQ_EPILOGUE();
}

void UART5_IRQHandler() // Motor 5
{
	CH_IRQ_PROLOGUE();
	UARTFifoIRQ(&motorsCom.motor[4].uart);
	CH_IRQ_EPILOGUE();
}

// ####################
// Public functions
// ####################
void MotorsComInit()
{
	motorsCom.state = STATE_NORMAL;
	motorsCom.driveCommutationMode = MOTOR_MODE_BLOCK;
	motorsCom.recordMotor = 0xFF;
	motorsCom.commKillMotor = 0xFF;

	chMBObjectInit(&motorsCom.eventQueue, motorsCom.pEventQueueData, MOTORS_COM_EVENT_QUEUE_SIZE);

	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_2MHZ;
	gpioInit.pupd = GPIO_PUPD_UP;

	// Motor 1 (FR)
	gpioInit.alternate = 7;
	GPIOInit(GPIOA, GPIO_PIN_9 | GPIO_PIN_10, &gpioInit);
	RCC->APB2ENR |= RCC_APB2ENR_USART1EN;

	// Motor 2 (FL)
	gpioInit.alternate = 7;
	GPIOInit(GPIOF, GPIO_PIN_6 | GPIO_PIN_7, &gpioInit);
	RCC->APB1LENR |= RCC_APB1LENR_UART7EN;

	// Motor 3 (RL)
	gpioInit.alternate = 8;
	GPIOInit(GPIOA, GPIO_PIN_0 | GPIO_PIN_1, &gpioInit);
	RCC->APB1LENR |= RCC_APB1LENR_UART4EN;

	// Motor 4 (RR)
	gpioInit.alternate = 7;
	GPIOInit(GPIOB, GPIO_PIN_10 | GPIO_PIN_11, &gpioInit);
	RCC->APB1LENR |= RCC_APB1LENR_USART3EN;

	// Motor 5 (Dribbler)
	gpioInit.alternate = 14;
	GPIOInit(GPIOB, GPIO_PIN_5 | GPIO_PIN_6, &gpioInit);
	RCC->APB1LENR |= RCC_APB1LENR_UART5EN;
	__DSB();

	UARTFifoInit(&motorsCom.motor[0].uart, USART1, 2000000, systemClockInfo.APB2PeriphClk, 1);
	UARTFifoInit(&motorsCom.motor[1].uart,  UART7, 2000000, systemClockInfo.APB1PeriphClk, 1);
	UARTFifoInit(&motorsCom.motor[2].uart,  UART4, 2000000, systemClockInfo.APB1PeriphClk, 1);
	UARTFifoInit(&motorsCom.motor[3].uart, USART3, 2000000, systemClockInfo.APB1PeriphClk, 1);
	UARTFifoInit(&motorsCom.motor[4].uart,  UART5, 2000000, systemClockInfo.APB1PeriphClk, 1);

	NVICEnableIRQ(USART1_IRQn, IRQL_MOTOR_COMMS);
	NVICEnableIRQ(UART7_IRQn, IRQL_MOTOR_COMMS);
	NVICEnableIRQ(UART4_IRQn, IRQL_MOTOR_COMMS);
	NVICEnableIRQ(USART3_IRQn, IRQL_MOTOR_COMMS);
	NVICEnableIRQ(UART5_IRQn, IRQL_MOTOR_COMMS);

	motorsCom.motor[0].portExTarget = PORT_EX_TARGET_MOTOR1;
	motorsCom.motor[1].portExTarget = PORT_EX_TARGET_MOTOR2;
	motorsCom.motor[2].portExTarget = PORT_EX_TARGET_MOTOR3;
	motorsCom.motor[3].portExTarget = PORT_EX_TARGET_MOTOR4;
	motorsCom.motor[4].portExTarget = PORT_EX_TARGET_DRIBBLER;

	MotorsComUpdateFirmware(0);
}

void MotorsComRunStatemachine()
{
	msg_t event;

	// process event if there is one
	if(motorsCom.state == STATE_NORMAL && chMBFetch(&motorsCom.eventQueue, &event, TIME_IMMEDIATE) == MSG_OK)
	{
		switch(event)
		{
			case EVENT_UPDATE_FIRMWARE:
			{
				motorsCom.state = STATE_BOOTLOADER;
				parallelUpdate(0);
//				for(uint8_t motor = 0; motor < 5; motor++)
//					flashProgram(motor, 0);
				motorsCom.state = STATE_NORMAL;
			}
			break;
			case EVENT_UPDATE_FIRMWARE_FORCE:
			{
				motorsCom.state = STATE_BOOTLOADER;
				parallelUpdate(1);
//				for(uint8_t motor = 0; motor < 5; motor++)
//					flashProgram(motor, 1);
				motorsCom.state = STATE_NORMAL;
			}
			break;
		}
	}
}

void MotorsComReceiveData()
{
	for(uint32_t motor = 0; motor < 5; motor++)
	{
		receiveMotorData(motor);
	}
}

void MotorsComStartSendData()
{
	for(uint32_t motor = 0; motor < 5; motor++)
	{
		sendMotorData(motor);
	}
}

void MotorsComWaitSendDataComplete()
{
	// wait for all motor transmissions to finish
	for(uint8_t motor = 0; motor < 5; motor++)
	{
		// need to finish communication in reverse order to make mutices happy
		MotorsSingle* pMotor = &motorsCom.motor[4-motor];

		// TODO: report failures to some statistics structure
		if(UARTFifoWriteWait(&pMotor->uart, MS2ST(10)))
			LogWarnC("Motor TX timeout", motor);
	}
}

void MotorsComUpdateFirmware(uint8_t force)
{
	if(force)
		chMBPost(&motorsCom.eventQueue, EVENT_UPDATE_FIRMWARE_FORCE, TIME_IMMEDIATE);
	else
		chMBPost(&motorsCom.eventQueue, EVENT_UPDATE_FIRMWARE, TIME_IMMEDIATE);
}

void MotorsComSetOff(uint8_t motor)
{
	motorsCom.motor[motor].mode = MODE_OFF;
}

void MotorsComSetVoltage(uint8_t motor, float voltage)
{
	motorsCom.motor[motor].mode = MODE_VOLTAGE;
	motorsCom.motor[motor].setVoltage = voltage;
}

void MotorsComSetCurrent(uint8_t motor, float current)
{
	motorsCom.motor[motor].mode = MODE_CURRENT;
	motorsCom.motor[motor].setCurrent = current;
}

void MotorsComSetPIGainsCurrent(uint8_t motor, float kp, float ki)
{
	motorsCom.motor[motor].curCtrlKp = FixedPointUnsignedSaturate(kp, 7, 10);
	motorsCom.motor[motor].curCtrlKi = FixedPointUnsignedSaturate(ki, 0, 18);
}

void MotorsComTestSetElectricalAngle(uint8_t motor, float angle, float voltage)
{
	motorsCom.motor[motor].mode = MODE_TEST_ANGLE;
	motorsCom.motor[motor].setAngle = angle;
	motorsCom.motor[motor].setVoltage = voltage;
}

void MotorsComDebugRecord(uint8_t id)
{
	recordCounter = 0;
	motorsCom.recordMotor = id;
}

void MotorsComDebugPrintRecording()
{
	for(uint16_t i = 0; i < RECORD_SIZE; i++)
	{
		ConsolePrint("%hu  %hd  %hd  %hd\r\n",
				recordVoltage[i], recordCurrent[i][0], recordCurrent[i][1], recordSpeed[i]);
		if(i % 10 == 0)
			chThdSleepMilliseconds(20);
	}
}

// ####################
// Private functions
// ####################
static void startBootloader(uint8_t target)
{
	PortExBootSet(target, 1);
	PortExResetPulse(target);
	PortExBootSet(target, 0);
}

static void flashProgram(uint16_t motorId, uint8_t force)
{
	if(motorId > 4)
		return;

	MotorsSingle* pMotor = &motorsCom.motor[motorId];

	STBootloaderInit(&pMotor->bootloader, &pMotor->uart, &startBootloader, pMotor->portExTarget);

	uint32_t originalBaudrate = pMotor->uart.baudrate;
	UARTFifoSetBaudrate(&pMotor->uart, 115200);

	pMotor->flashFuncResult = STBootloaderFlash(&pMotor->bootloader, (uint8_t*)MOTORS_SOURCE_ADDRESS,
			MOTORS_PROGRAM_SIZE, MOTORS_TARGET_ADDRESS, force, &pMotor->flashResult);

	UARTFifoSetBaudrate(&pMotor->uart, originalBaudrate);
}

static void flashTask(void* params)
{
	chRegSetThreadName("FlashMotor");

	uint32_t allParams = *((uint32_t*)params);
	flashProgram(allParams & 0xFFFF, allParams >> 16);

	chThdExit(0);
}

static void parallelUpdate(uint32_t force)
{
	for(uint32_t i = 0; i < 5; i++)
	{
		MotorsSingle* pMotor = &motorsCom.motor[i];

		pMotor->flashParam = force << 16 | i;
		pMotor->pFlashTask = chThdCreateStatic(pMotor->waFlashTask, sizeof(pMotor->waFlashTask), LOWPRIO,
				&flashTask, &pMotor->flashParam);
	}

	for(uint32_t i = 0; i < 5; i++)
	{
		MotorsSingle* pMotor = &motorsCom.motor[i];

		chThdWait(pMotor->pFlashTask);
	}

	ConsolePrint("    Req  Upd  Res     Size    Blocks  Time\r\n");
	for(uint16_t i = 0; i < 5; i++)
	{
		MotorsSingle* pMotor = &motorsCom.motor[i];
		ConsolePrint("M%hu  %hu    %hu    0x%04hX  %-6u%4u     %ums\r\n", i,
				(uint16_t)pMotor->flashResult.updateRequired, (uint16_t)pMotor->flashResult.updated,
				pMotor->flashFuncResult, pMotor->flashResult.programSize, pMotor->flashResult.numBlocks,
				pMotor->flashResult.timeUs/1000);
	}
}

static void receiveMotorData(uint32_t motorId)
{
	MotorsSingle* pMotor = &motorsCom.motor[motorId];

	int16_t result;
	uint16_t fifoStatus = 0;
	uint32_t rxTime;
	uint16_t datasize;

	while(1)
	{
		// Read UART data
		datasize = sizeof(MotorExchangeMISO);
		result = UARTFifoRead(&pMotor->uart, (uint8_t*)&pMotor->miso, &datasize, &fifoStatus, &rxTime);
		if(result == UART_FIFO_ERROR_NO_DATA)
			break;
		if(result)
		{
			LogWarnC("read error", result | motorId << 8 | (uint32_t)fifoStatus << 16);
			continue;
		}
		if(datasize < sizeof(MotorExchangeMISO))
		{
			LogWarnC("Short read", datasize | motorId << 8);
			continue;
		}

		// process CPU load
		pMotor->rxTime = rxTime;
		pMotor->cpuLoad = ((float)(pMotor->miso.idleTicksNoLoad - pMotor->miso.idleTicksActive))/((float)pMotor->miso.idleTicksNoLoad)*100.0f;

		// process voltage, current, and temperature measurement
		pMotor->avgVoltage1ms = pMotor->miso.avgVoltage * 0.001f;
		pMotor->avgTemperature1ms = pMotor->miso.avgTemperature * 0.1f;

		float cur[2] = { 0.0f, 0.0f };
		for(uint8_t i = 0; i < MOTOR_EXCHANGE_MISO_MAX_MEAS; i++)
		{
			cur[0] += pMotor->miso.currentMeas[i][0];
			cur[1] += pMotor->miso.currentMeas[i][1];
		}

		const float factor = 1.0f/((float)MOTOR_EXCHANGE_MISO_MAX_MEAS) * 0.001f;
		pMotor->avgCurrent1ms = cur[0] * factor;
		pMotor->avgCurrent1ms2 = cur[1] * factor;

		pMotor->currentBias = pMotor->miso.currentBias * 0.001f;

		// process speed measurements
		pMotor->encoderVelocity = pMotor->miso.encDelta * HALL_SPEED_TO_RAD_PER_SEC_DRIVE;

		if(motorId > 3)
			pMotor->hallVelocity = pMotor->miso.hallSpeed * HALL_SPEED_TO_RAD_PER_SEC_DRIBBLER;
		else
			pMotor->hallVelocity = pMotor->miso.hallSpeed * HALL_SPEED_TO_RAD_PER_SEC_DRIVE;

		// record raw data if requested
		if(motorsCom.recordMotor == motorId)
		{
			if(recordCounter+MOTOR_EXCHANGE_MISO_MAX_MEAS < RECORD_SIZE)
			{
				memcpy(&recordCurrent[recordCounter], pMotor->miso.currentMeas, sizeof(int16_t)*MOTOR_EXCHANGE_MISO_MAX_MEAS*2);

				for(uint8_t i = 0; i < MOTOR_EXCHANGE_MISO_MAX_MEAS; i++)
				{
					recordVoltage[recordCounter+i] = pMotor->miso.avgVoltage;
					recordSpeed[recordCounter+i] = pMotor->miso.encDelta;
				}

				recordCounter += MOTOR_EXCHANGE_MISO_MAX_MEAS;
			}
			else
			{
				motorsCom.recordMotor = 0xFF;
			}
		}
	}
}

static void sendMotorData(uint32_t motorId)
{
	if(motorsCom.commKillMotor == motorId)
		return;

	MotorsSingle* pMotor = &motorsCom.motor[motorId];
	MotorExchangeMOSI* pMosi = &pMotor->mosi;

	pMosi->curCtrlKp = FixedPointSignedFromFloat(pMotor->curCtrlKp, 10);
	pMosi->curCtrlKi = FixedPointSignedFromFloat(pMotor->curCtrlKi, 18);

	pMosi->flags = 0;

	if(pMotor->miso.flags & MOTOR_EXCHANGE_MISO_FLAG_BREAK_ACTIVE)
	{
		BuzzerPlay(&buzzSeqUp20);
		pMosi->flags |= MOTOR_EXCHANGE_MOSI_FLAG_CLEAR_BREAK;
		LogWarnC("Active break on motor", motorId);
	}

	switch(pMotor->mode)
	{
		case MODE_VOLTAGE:
		{
			pMosi->motorMode = motorsCom.driveCommutationMode;
			pMosi->ctrlMode = MOTOR_CTRL_NONE;
			pMosi->setVoltage = (int32_t)(pMotor->setVoltage*1000.0f);
		}
		break;
		case MODE_CURRENT:
		{
			pMosi->motorMode = motorsCom.driveCommutationMode;
			pMosi->ctrlMode = MOTOR_CTRL_CURRENT;
			pMosi->setCurrent = (int16_t)(pMotor->setCurrent*1000.0f);
		}
		break;
		case MODE_TEST_ANGLE:
		{
			pMosi->motorMode = MOTOR_MODE_ANGLE_TEST;
			pMosi->ctrlMode = MOTOR_CTRL_NONE;
			pMosi->setVoltage = (int32_t)(pMotor->setVoltage*1000.0f);
			pMosi->angle = (int32_t)(AngleNormalize(pMotor->setAngle) * 4096.0f/(float)M_PI) + 4096;
		}
		break;
		default:
		{
			pMosi->motorMode = MOTOR_MODE_OFF;
		}
		break;
	}

	pMotor->txTime = SysTimeUSec();

	UARTFifoWriteStart(&pMotor->uart, (uint8_t*)&pMotor->mosi, sizeof(MotorExchangeMOSI));
}
