/*
 * presenter.c
 *
 *  Created on: 09.01.2019
 *      Author: AndreR
 */

#include "presenter.h"
#include "hal/tft.h"
#include "gfx.h"
#include "util/console.h"
#include "main/network.h"
#include "power.h"
#include "util/log_file.h"
#include "util/network_print.h"
#include "prog_source_usb.h"
#include "main/robot.h"
#include "hal/kicker.h"
#include "motors_com.h"
#include "ir.h"

#include "gui/bot_cover.h"
#include "gui/main_status.h"
#include "gui/top_bar.h"
#include "gui/menu.h"
#include "gui/wifi.h"
#include "gui/select_botid.h"
#include "gui/ids.h"
#include "gui/logging.h"
#include "gui/test_drive.h"
#include "gui/versions.h"

#define GUI_HEAP_SIZE (20*1024)
memory_heap_t guiHeap;
uint8_t guiHeapBuffer[GUI_HEAP_SIZE] __attribute__((aligned(sizeof(stkalign_t))));

typedef int16_t(*EventHandlerFunc)(GEvent*);

static struct _handles
{
	GHandle hTopBar;
	GHandle hMenu;

	GHandle hWindows[9];
} handles;

static int16_t currentWindow = GUI_WINDOW_STATUS;


static void logFileCallback(LogFileEvent* pLogEvent)
{
	LoggingLogUpdate(pLogEvent);

	if(pLogEvent->eventType == LOGFILE_EVENT_START)
	{
		ConsolePrint("Logfile started: %s\r\n", logFile.logFilename);
		NetworkPrint("Logfile started: %s\n", logFile.logFilename);
	}

	if(pLogEvent->eventType == LOGFILE_EVENT_CLOSE)
	{
		ConsolePrint("Log file closed\r\n");
		NetworkPrint("Log file closed\n");
	}
}

static void usbHcdCallback(USBHCDConnectionEvent* pConEvent)
{
	LoggingHCDUpdate(pConEvent);
}

static void usbMscCallback(USBMSCEvent* pMscEvent)
{
	LoggingMSCUpdate(pMscEvent);

	if(pMscEvent->connect)
	{
		if(f_stat("autorun_bootloader", 0) == FR_OK)
			ProgStoreStart(PROG_STORE_SOURCE_USB);
	}
}

void PresenterInit()
{
	chHeapObjectInit(&guiHeap, guiHeapBuffer, GUI_HEAP_SIZE);

	logFile.eventCb = &logFileCallback;
	usbHcd.pConCb = &usbHcdCallback;
	usbMsc.eventCb = &usbMscCallback;
}

void PresenterShowWindow(int16_t newWindow)
{
	if(newWindow < 0 || newWindow > 9)
		return;

	gwinSetVisible(handles.hWindows[currentWindow], FALSE);

	currentWindow = newWindow;

	TopBarSetTitle(((GWidgetObject*)handles.hWindows[currentWindow])->text);

	gwinSetVisible(handles.hMenu, FALSE);
	gwinSetVisible(handles.hTopBar, TRUE);
	gwinSetVisible(handles.hWindows[currentWindow], TRUE);
}


void PresenterTask(void* params)
{
	(void)params;

	chRegSetThreadName("Pres");

	gfxInit();
	gdispSetOrientation(GDISP_ROTATE_180);

	// Set the widget defaults
	gwinSetDefaultFont(gdispOpenFont("fixed_7x14"));
	gwinSetDefaultStyle(&BlackWidgetStyle, FALSE);
	gwinSetDefaultBgColor(Black);
	gwinSetDefaultColor(White);
	gdispClear(Black);

	handles.hTopBar = TopBarCreate();
	handles.hWindows[GUI_WINDOW_STATUS] = MainStatusCreate();
	handles.hWindows[GUI_WINDOW_WIFI_SETTINGS] = WifiCreate();
	handles.hWindows[GUI_WINDOW_BOT_ID] = SelectBotIDCreate();
	handles.hWindows[GUI_WINDOW_LOGGING] = LoggingCreate();
	handles.hWindows[GUI_WINDOW_TEST_DRIVE] = TestDriveCreate();
	handles.hWindows[GUI_WINDOW_VERSIONS] = VersionsCreate();

	handles.hMenu = MenuCreate(handles.hWindows, 9);

	GListener gl;

	geventListenerInit(&gl);
	gwinAttachListener(&gl);

	TopBarSetTitle(((GWidgetObject*)handles.hWindows[currentWindow])->text);
	gwinSetVisible(handles.hWindows[currentWindow], TRUE);

	while(1)
	{
		GEvent* pEvent;

		pEvent = geventEventWait(&gl, 40);

		if(pEvent == 0)
		{
			MainStatusWifiUpdate(&network.config);

			KdFeedback kd;
			kd.barrier.level[0] = kicker.vIrOn*1000.0f;
			kd.barrier.level[1] = kicker.vIrOff*1000.0f;
			kd.barrier.threshold = kicker.irAutoThreshold*1000.0f;
//			kd.barrier.threshold = kicker.config.irThreshold*1000.0f;
			kd.power.bat = power.vBat;
			kd.power.cur = power.iCur;
			kd.kicker.cap = kicker.vCap;
			kd.kicker.chg = 0.0f;
			kd.kicker.temp = kicker.tEnv;
			MainStatusKdUpdate(&kd);
			MainStatusHwIdUpdate(robot.misc.hardwareId);

			// Wifi
			WifiUpdateConfig(&network.config);
			WifiUpdateStatus(&robot.wifiStat);

			VersionsUpdateMotor(0, &motorsCom.motor[0].flashResult);
			VersionsUpdateIr(&ir.flashResult);

			continue;
		}

		if(TopBarIsMenuClicked(pEvent))
		{
			gwinSetVisible(handles.hTopBar, FALSE);
			gwinSetVisible(handles.hWindows[currentWindow], FALSE);

			gwinSetVisible(handles.hMenu, TRUE);

			continue;
		}

		int16_t menuSelection = -1;
		if(((GWidgetObject*)handles.hWindows[currentWindow])->fnParam)
		{
			// please close your eyes for the next line, looks messy ;)
			menuSelection = ((EventHandlerFunc)(((GWidgetObject*)handles.hWindows[currentWindow])->fnParam))(pEvent);
		}

		if(menuSelection < 0)
			menuSelection = MenuGetSelection(pEvent);

		if(menuSelection >= 0)
		{
			PresenterShowWindow(menuSelection);
		}
	}
}

void PresenterPrintHeapStatus()
{
	size_t free;
	size_t fragments = chHeapStatus(&guiHeap, &free);

	ConsolePrint("Fragments: %u, Free: %u, Core: %u\r\n", fragments, free, chCoreGetStatusX());
}
