/*
 * prog_source_usb.c
 *
 *  Created on: 25.06.2019
 *      Author: AndreR
 */

#include "prog_source_usb.h"
#include "util/crc.h"
#include "util/console.h"

#include <string.h>

static void dataFunc(uint8_t procId, uint32_t offset, uint32_t size);
static void crcFunc(uint8_t procId, uint32_t startAddr, uint32_t endAddr);
static void sizeFunc(uint8_t procId);
static void open();
static void close();

ProgSourceUSB progSourceUSB = {.usb = {&dataFunc, &sizeFunc, &crcFunc, &open, &close, 100, PROG_SOURCE_USB_BUF}};

const char* mainFilename = "0:/main2019.bin";
const char* irFilename = "0:/ir2019.bin";
const char* motorFilename = "0:/motor2019.bin";

void ProgSourceUSBInit()
{
	ProgStoreAddSource(&progSourceUSB.usb);
}

static void open()
{
	f_open(&progSourceUSB.mainFile, (TCHAR*)mainFilename, FA_READ | FA_OPEN_EXISTING);
	f_open(&progSourceUSB.irFile, (TCHAR*)irFilename, FA_READ | FA_OPEN_EXISTING);
	f_open(&progSourceUSB.motorFile, (TCHAR*)motorFilename, FA_READ | FA_OPEN_EXISTING);
}

static void close()
{
	f_close(&progSourceUSB.mainFile);
	f_close(&progSourceUSB.irFile);
	f_close(&progSourceUSB.motorFile);
}

static void dataFunc(uint8_t procId, uint32_t offset, uint32_t size)
{
	FIL* pFile;

	switch(procId)
	{
		case PROG_STORE_ID_MAIN2019:
			pFile = &progSourceUSB.mainFile;
			break;
		case PROG_STORE_ID_IR2019:
			pFile = &progSourceUSB.irFile;
			break;
		case PROG_STORE_ID_MOT2019:
			pFile = &progSourceUSB.motorFile;
			break;
		default:
			ProgStoreProvideError(procId, PROG_STORE_ERROR_NO_MEDIA);
			return;
	}

	memset(progSourceUSB.buf, 0xFF, PROG_SOURCE_USB_BUF);

	UINT bytesRead;
	FRESULT fresult;
	fresult = f_lseek(pFile, offset);
	if(fresult)
	{
		ProgStoreProvideError(procId, PROG_STORE_ERROR_NO_MEDIA);
		return;
	}

	fresult = f_read(pFile, progSourceUSB.buf, PROG_SOURCE_USB_BUF, &bytesRead);
	if(fresult)
	{
		ProgStoreProvideError(procId, PROG_STORE_ERROR_NO_MEDIA);
		return;
	}

	ProgStoreProvideData(procId, offset, progSourceUSB.buf, size);
}

static void sizeFunc(uint8_t procId)
{
	const char* pFilename;

	switch(procId)
	{
		case PROG_STORE_ID_MAIN2019:
			pFilename = mainFilename;
			break;
		case PROG_STORE_ID_IR2019:
			pFilename = irFilename;
			break;
		case PROG_STORE_ID_MOT2019:
			pFilename = motorFilename;
			break;
		default:
			ProgStoreProvideError(procId, PROG_STORE_ERROR_NO_MEDIA);
			return;
	}

	FRESULT fresult;
	FILINFO info;
	memset(&info, 0, sizeof(FILINFO));

	fresult = f_stat(pFilename, &info);
	if(fresult)
	{
		ProgStoreProvideError(procId, PROG_STORE_ERROR_NO_MEDIA);
		return;
	}

	ProgStoreProvideSize(procId, info.fsize);
}

static void crcFunc(uint8_t procId, uint32_t startAddr, uint32_t endAddr)
{
	FIL* pFile;

	switch(procId)
	{
		case PROG_STORE_ID_MAIN2019:
			pFile = &progSourceUSB.mainFile;
			break;
		case PROG_STORE_ID_IR2019:
			pFile = &progSourceUSB.irFile;
			break;
		case PROG_STORE_ID_MOT2019:
			pFile = &progSourceUSB.motorFile;
			break;
		default:
			ProgStoreProvideError(procId, PROG_STORE_ERROR_NO_MEDIA);
			return;
	}

	UINT bytesRead;
	FRESULT fresult;
	fresult = f_lseek(pFile, startAddr);
	if(fresult)
	{
		ProgStoreProvideError(procId, PROG_STORE_ERROR_NO_MEDIA);
		return;
	}

	uint32_t addr = 0;
	uint32_t size = endAddr - startAddr;

	uint32_t crc = CRC32StartChecksumAlt();

	while(addr < size)
	{
		uint32_t readSize = PROG_SOURCE_USB_BUF;

		if(addr + PROG_SOURCE_USB_BUF > size)
			readSize = size-addr;

		memset(progSourceUSB.buf, 0xFF, PROG_SOURCE_USB_BUF);

		fresult = f_read(pFile, progSourceUSB.buf, readSize, &bytesRead);
		if(fresult)
		{
			ProgStoreProvideError(procId, PROG_STORE_ERROR_NO_MEDIA);
			break;
		}

		crc = CRC32FeedChecksumAlt(crc, progSourceUSB.buf, readSize);

		addr += readSize;
	}

	crc = CRC32StopChecksumAlt(crc);

	ProgStoreProvideCRC32(procId, startAddr, endAddr, crc);
}
