/*
 * robot_impl.c
 *
 *  Created on: 12.01.2019
 *      Author: AndreR
 */

#include "main/robot.h"
#include "main/ctrl.h"

#include "hal/kicker.h"
#include "hal/buzzer.h"
#include "hal/led.h"

#include "util/sys_time.h"
#include "util/log_file.h"

#include "power.h"
#include "spi4.h"
#include "motors.h"
#include "motors_com.h"
#include "ext.h"

BotParameters botParams = {
	.driveTrain = {
		.motor2WheelRatio = (30.0f/50.0f),
		.wheel2MotorRatio = (50.0f/30.0f),
		.motor = {
			.Km = 0.0369f,
			.Kn = (259.0f*2.0f*PI/60.0f),
			.Ke = (60.0f/(259.0f*2.0f*PI)),
			.Kf = 0.94e-4f,
			.R = 0.608f,
			.L = 463e-6f,
		}
	},
	.physical = { 0.0165f, 45.0f, 45.0f, 0.082f, 2.46f }, // neu: 2.46kg,	Alt: 2.65kg
	.v2016 = 0,
};

void RobotImplUpdateSensors(RobotSensors* pSensors)
{
	// motors
	memset(&pSensors->enc, 0, sizeof(pSensors->enc));
	pSensors->enc.time = SysTimeUSec();
	pSensors->motorVol.time = SysTimeUSec();

	for(uint8_t m = 0; m < 4; m++)
	{
		pSensors->enc.vel[m] = motorsCom.motor[m].encoderVelocity;
		pSensors->motorVol.vol[m] = motorsCom.motor[m].setVoltage;
	}

	// motor current
	memset(&pSensors->motorCur, 0, sizeof(pSensors->motorCur));
	pSensors->motorCur.time = SysTimeUSec();
	pSensors->motorCur.cur[0] = motorsCom.motor[0].avgCurrent1ms;
	pSensors->motorCur.cur[1] = motorsCom.motor[1].avgCurrent1ms;
	pSensors->motorCur.cur[2] = motorsCom.motor[2].avgCurrent1ms;
	pSensors->motorCur.cur[3] = motorsCom.motor[3].avgCurrent1ms;

	// power
	pSensors->power.voltage = power.vBat;
	pSensors->power.current = power.iCur;
	pSensors->power.batEmpty = power.batEmpty;
	pSensors->power.exhausted = power.exhausted;

	// IMU
	pSensors->acc.linAcc[0] = spi4.imu.acc[0];
	pSensors->acc.linAcc[1] = spi4.imu.acc[1];
	pSensors->acc.linAcc[2] = spi4.imu.acc[2];
	pSensors->acc.time = SysTimeUSec();
	pSensors->acc.updated = 1;

	pSensors->gyr.rotVel[0] = spi4.imu.gyr[0];
	pSensors->gyr.rotVel[1] = spi4.imu.gyr[1];
	pSensors->gyr.rotVel[2] = spi4.imu.gyr[2];
	pSensors->gyr.time = SysTimeUSec();
	pSensors->gyr.updated = 1;

	// Vision - SKIPPED (filled by robot base implementation)

	// Kicker
	pSensors->kicker.level = kicker.vCap;
	pSensors->kicker.chgCurrent = 0.0f;
	pSensors->kicker.kickCounter = kicker.kickCounter;
	pSensors->kicker.flags = 0;
	if(KickerIsReadyToShoot())
		pSensors->kicker.flags |= KICKER_FLAGS_READY;

	if(kicker.barrierDamaged)
		pSensors->kicker.flags |= KICKER_FLAGS_DMG_BARRIER;

	if(kicker.straightDamaged)
		pSensors->kicker.flags |= KICKER_FLAGS_DMG_STRAIGHT;

	if(kicker.chipDamaged)
		pSensors->kicker.flags |= KICKER_FLAGS_DMG_CHIP;

	if(kicker.chargeOverheated)
		pSensors->kicker.flags |= KICKER_FLAGS_DMG_CHARGE;

	if(kicker.v2017)
		pSensors->kicker.flags |= KICKER_FLAGS_V2017;

	// Dribbler
	pSensors->dribbler.speed = motorsCom.motor[4].hallVelocity*MOTORS_DRIBBLER_MOTOR_RADS_TO_BAR_RPM;
	pSensors->dribbler.voltage = motorsCom.motor[4].setVoltage;
	pSensors->dribbler.temp = motors.dribbler.temperature;
	pSensors->dribbler.overheated = motors.dribbler.overheated;

	// Barrier
	pSensors->ir.level[0] = kicker.vIrOff;
	pSensors->ir.level[1] = kicker.vIrOn;
	pSensors->ir.interrupted = kicker.interrupted;
}

void RobotImplWriteLogMsg(void* pMsg)
{
	LogFileWrite(pMsg);
}

void RobotImplMotorOutput(const RobotCtrlOutput* pMotor)
{
	motors.driveMode = pMotor->ctrlMode;

	for(uint8_t i = 0; i < 4; i++)
	{
		motors.drive[i].setVelocity = pMotor->motorVel[i];
		motors.drive[i].setVoltage = pMotor->motorVoltage[i];
		motors.drive[i].setAcceleration = pMotor->motorAcc[i];
		motors.drive[i].setTorque = pMotor->motorTorque[i];
	}
}

void RobotImplKickerOutput(const KickerInput* pKicker)
{
	switch(pKicker->mode)
	{
	case KICKER_MODE_FORCE:
		KickerFireSpeed(pKicker->speed, pKicker->device);
		break;
	case KICKER_MODE_ARM:
		KickerArmSpeed(pKicker->speed, pKicker->device);
		break;
	case KICKER_MODE_DISARM:
		KickerDisarm();
		break;
	case KICKER_MODE_ARM_TIME:
		KickerArm(pKicker->speed, pKicker->device);
		break;
	}
}

void RobotImplDribblerOutput(const DribblerInput* pDribbler)
{
	switch(pDribbler->mode)
	{
	case DRIBBLER_MODE_VOLTAGE:
		MotorsDribblerSetVoltage(pDribbler->voltage);
		break;
	case DRIBBLER_MODE_SPEED:
		MotorsDribblerSetVelocity(pDribbler->speed);
		break;
	default:
		MotorsDribblerOff();
		break;
	}
}

void RobotImplBuzzerPlay(uint16_t seqId)
{
	BuzzerPlayId(seqId);
}

void RobotImplKickerAutoCharge(uint8_t enable)
{
	if(enable)
	{
		KickerEnableAutoRecharge(1);
		KickerEnableCharge(1);
	}
	else
	{
		KickerAutoDischarge();
	}
}

void RobotImplSetLEDs(uint8_t state)
{
	switch(state)
	{
		case SYSTEM_MATCH_CTRL_FLAGS_LED_OFF:
			LEDLeftSet(0, 0, 0, 0.01);
			LEDRightSet(0, 0, 0, 0.01);
			break;
		case SYSTEM_MATCH_CTRL_FLAGS_LED_RED:
			LEDLeftSet(0.5f, 0, 0, 0);
			LEDRightSet(0.5f, 0, 0, 0);
			break;
		case SYSTEM_MATCH_CTRL_FLAGS_LED_GREEN:
			LEDLeftSet(0, 0.3f, 0, 0);
			LEDRightSet(0, 0.3f, 0, 0);
			break;
		case SYSTEM_MATCH_CTRL_FLAGS_LED_BLUE:
			LEDLeftSet(0, 0, 0.4f, 0);
			LEDRightSet(0, 0, 0.4f, 0);
			break;
		case SYSTEM_MATCH_CTRL_FLAGS_LED_WHITE:
			LEDLeftSet(0, 0, 0, 0.5f);
			LEDRightSet(0, 0, 0, 0.5f);
			break;
		case SYSTEM_MATCH_CTRL_FLAGS_LED_LIGHT_BLUE:
			LEDLeftSet(0, 0.25f, 0.25f, 0);
			LEDRightSet(0, 0.25f, 0.25f, 0);
			break;
		case SYSTEM_MATCH_CTRL_FLAGS_LED_GOLD:
			LEDLeftSet(0.3f, 0.15f, 0, 0);
			LEDRightSet(0.3f, 0.15f, 0, 0);
			break;
		case SYSTEM_MATCH_CTRL_FLAGS_LED_PURPLE:
			LEDLeftSet(0.25f, 0, 0.25f, 0);
			LEDRightSet(0.25f, 0, 0.25f, 0);
			break;
		default:
			LEDLeftSet(0, 0.1f, 0, 0);
			LEDRightSet(0, 0, 0.1f, 0);
			break;
	}
}

void RobotImplExtSendMatchFeedback(const SystemMatchFeedback* pFeedback)
{
	static uint32_t lastFeedbackTime = 0;

	if((SysTimeUSecRaw() - lastFeedbackTime) < 10000)
		return;

	lastFeedbackTime = SysTimeUSecRaw();

	PacketHeader header;
	header.section = SECTION_SYSTEM;
	header.cmd = CMD_SYSTEM_MATCH_FEEDBACK;
	ExtSendPacket(&header, pFeedback, sizeof(SystemMatchFeedback));
}
