/*
 * adc.c
 *
 *  Created on: 18.02.2019
 *      Author: AndreR
 */

/**
 * Motor current calculations to ADC:
Vdd :3.3 $
Rs :0.025 $
Rb :16.2e3 $
Rlp :1.33e3 $
U_signal:I*Rs*Rb/(Rb+Rlp);
U_bias:Vdd*Rlp/(Rb+Rlp);
U:U_signal+ U_bias;
A:(7.32+1.33)/1.33;
U_adc =expand(A*U*2^12/3.3);
s1 :solve(U_adc =expand(A*U*2^12/3.3), I);
float(expand(s1[1]));

I [mA] = 5.361871705068514*ADC_VAL - 10837.03703703705
I = m*ADC + b
ADC: Q12.0 => 0 - 4096
m: Q3.15 => 5.36187744140625 (error -0.000005736337735984254) (5 << 15 | 11858)
b: Q14.15 => 10837.03704833984375 (error -0.00001130279375000098) (10837 << 15 | 1214)
m*ADC => Q15.15 + b => Q16.15

 * Voltage measurement:
R1 :243e3$
R2 :24.1e3$
(R1+R2)/R2;
ADC=U_in *R2/(R1+R2)*2^12/3.3;
s1 :solve(ADC=U_in *R2/(R1+R2)*2^12/3.3, U_in);
float(s1[1])*1000;

U [mV] = 8.929164775674275*ADC_VAL
U = m*ADC
ADC: Q12.0 => 0 - 4096
m: Q4.15 => 8.929168701171875 (error -0.000003925497599976957) (8 << 15 | 30447)
m*ADC => Q16.15

 * Temperature measurement:
ad(t):=(1.43 + (t-30)*4.3e-3)*4096/3.3;
ts1 :ad(30);
ts2:ad(110);
tsd:ad(40);
(110-30)/(ts2-ts1)*(tsd-ts1) + 30;
factor :floor(1/(ts2-ts1)*(2^29-1) * 800 * 1/2^12);
floor(f*4096);
floor(%) * 1/2^17.0 + 300;
floor(1/2^7*(2^29-1) * 800);

 * ADC calculations:
t_adc :12.5$
t_smp :1.5 $
t_latency : 2.625 $
t_trig : 0.219e-6 $
f_adc :12e6 $
t_conv :(t_adc+t_smp+t_latency)/f_adc + t_trig;
t_total :2*t_conv;
 */

#include "adc.h"
#include "usart1.h"
#include "system_init.h"
#include "log_msgs.h"
#include "motor.h"
#include "ctrl.h"
#include "data.h"
#include <string.h>

ADCData adc;

static void doControl();
static void startTemperatureMeasurementAndDoControl();
static void configureCurrentMeasurement();

void DMA1_Channel1_IRQHandler()
{
	// clear flags
	DMA1->IFCR = DMA_IFCR_CGIF1;
	ADC1->ISR = ADC_ISR_EOSMP | ADC_ISR_EOC | ADC_ISR_EOSEQ | ADC_ISR_OVR;
	
	// configure for voltage measurement
	ADC1->CHSELR = ADC_CHSELR_CHSEL3; // bus voltage
	ADC1->CFGR1 = ADC_CFGR1_OVRMOD;
	ADC1->IER = ADC_IER_EOSEQIE;
	ADC1->CR |= ADC_CR_ADSTART;

	adc.pNextAdcFunc = &startTemperatureMeasurementAndDoControl;
}

void ADC1_IRQHandler()
{
	// clear flags
	ADC1->ISR = ADC_ISR_EOSMP | ADC_ISR_EOC | ADC_ISR_EOSEQ | ADC_ISR_OVR;

	(*adc.pNextAdcFunc)();
}

static void startTemperatureMeasurementAndDoControl()
{
	adc.volAdc_U12_0 = ADC1->DR;

	ADC1->CHSELR = ADC_CHSELR_CHSEL16; // temperature sensor
	ADC1->SMPR = ADC_SMPR_SMP_2 | ADC_SMPR_SMP_1; // 71.5 clock cycles sampling time
	ADC1->CFGR1 = ADC_CFGR1_OVRMOD;
	ADC1->CR |= ADC_CR_ADSTART;

	adc.pNextAdcFunc = &configureCurrentMeasurement;

	doControl();
}

static void configureCurrentMeasurement()
{
	adc.tempAdc_S12_0 = ADC1->DR;

	ADC1->CHSELR = ADC_CHSELR_CHSEL4; // channel 4 (motor current)
	ADC1->SMPR = 0; // 1.5 ADC clock cycles sampling time
	ADC1->CFGR1 = ADC_CFGR1_OVRMOD | ADC_CFGR1_DMAEN | ADC_CFGR1_EXTSEL_0 | ADC_CFGR1_EXTEN;
	ADC1->IER = 0;
	ADC1->CR |= ADC_CR_ADSTART;
}

static inline void doControl()
{
	const int32_t curMult_Q3_15 = 175698L; // 5.361871705068514
	const int32_t curOffset_Q14_15 = 355108030L; // 10837.03703703705
	const uint32_t volMult_Q4_15 = 292591UL; // 8.929164775674275

	int32_t curSine0Adc_Q12_0 = (int32_t)adc.dmaData[1];
	int32_t curSine1Adc_Q12_0 = (int32_t)adc.dmaData[2];
	// TODO: below a certain modulation index dmaData[0] could be used as offset measurement

	int32_t currentSineHigh_Q14_15 = curMult_Q3_15 * curSine0Adc_Q12_0 - curOffset_Q14_15;
	int32_t currentSineMid_Q14_15 = curMult_Q3_15 * curSine1Adc_Q12_0 - curOffset_Q14_15;

	int32_t factorHigh_S0_12 = motor.VHighUsed_U12_0 - motor.VMidUsed_U12_0;
	int32_t factorMid_S0_12 = motor.VMidUsed_U12_0 - motor.VLowUsed_U12_0;

	int32_t dcCurrentHigh_S14_0 = (((currentSineHigh_Q14_15 >> 15) * factorHigh_S0_12) >> 12);
	int32_t dcCurrentMid_S14_0 = (((currentSineMid_Q14_15 >> 15) * factorMid_S0_12) >> 12);

	int32_t dcCurrent_S14_0 = dcCurrentHigh_S14_0 + dcCurrentMid_S14_0;
	adc.pSampleCurrent->iMotor = dcCurrent_S14_0;
	adc.pSampleCurrent->iMotor2 = 0;

	CtrlUpdatePerPWMCycle(currentSineHigh_Q14_15, currentSineMid_Q14_15);

	adc.currentSum += dcCurrent_S14_0;

	// resulting voltage is in [mV] and Q16.15 fixed-point representation
	uint32_t voltage_Q16_15 = volMult_Q4_15 * adc.volAdc_U12_0;

	adc.voltageSum += voltage_Q16_15 >> 15;

	// resulting temperature is in dezi-degree celsius [dC]
	int32_t temperature_Q15_0 = (((adc.tempAdc_S12_0 - adc.temperatureOffset) * adc.temperatureFactor) >> 17) + 300;

	adc.tempSum += temperature_Q15_0;

	++adc.pSampleCurrent;

	if(adc.pSampleCurrent == adc.pSampleBufEnd)
	{
		// all samples in a 1ms interval taken, swap active buffers
		// and trigger low-priority task (main)
		if(adc.pSampleBufEnd == &adc.sampleBuf[0][ADC_NUM_SAMPLES_PER_MS])
		{
			adc.pSampleBufEnd = &adc.sampleBuf[1][ADC_NUM_SAMPLES_PER_MS];
			adc.pSampleCurrent = adc.sampleBuf[1];
			adc.pFinishedSampleBuf = adc.sampleBuf[0];
		}
		else
		{
			adc.pSampleBufEnd = &adc.sampleBuf[0][ADC_NUM_SAMPLES_PER_MS];
			adc.pSampleCurrent = adc.sampleBuf[0];
			adc.pFinishedSampleBuf = adc.sampleBuf[1];
		}

		adc.avgVoltage_Q16_0 = (adc.voltageSum*205) >> 12; // => *1/20
		adc.avgCurrent_S14_0 = (adc.currentSum*205) >> 12;
		adc.avgTemperature_S15_0 = (adc.tempSum*205) >> 12;

		adc.voltageSum = 0;
		adc.currentSum = 0;
		adc.tempSum = 0;

		adc.trigger1ms = 1;
	}
}

void ADCInit()
{
	adc.avgVoltage_Q16_0 = 22200;

	adc.pSampleCurrent = adc.sampleBuf[0];
	adc.pSampleBufEnd = &adc.sampleBuf[0][ADC_NUM_SAMPLES_PER_MS];
	adc.pFinishedSampleBuf = adc.sampleBuf[1];

	// internal temperature sensor calibration values at 30C and 110C
	int32_t tsCal30Deg = *((volatile uint16_t*)0x1FFFF7B8);
	int32_t tsCal110Deg = *((volatile uint16_t*)0x1FFFF7C2);

	// ((2^29-1) / (ts2-ts1)) * 800 * 1/2^12
	adc.temperatureFactor = (0x1FFFFFFF/(tsCal110Deg - tsCal30Deg) * 800L) >> 12;
	adc.temperatureOffset = tsCal30Deg;

	RCC->APB2ENR |= RCC_APB2ENR_ADCEN;

	// select 12MHz PCLK as source
	ADC1->CFGR2 = ADC_CFGR2_JITOFFDIV4;

	// do ADC calibration
	ADC1->CR = ADC_CR_ADCAL;
	while(ADC1->CR & ADC_CR_ADCAL);

	// configure analog input pins
	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_ANALOG;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOA, GPIO_PIN_3 | GPIO_PIN_4, &gpioInit);

	ADC1->ISR = ADC_ISR_ADRDY;
	ADC1->CR = ADC_CR_ADEN;
	while((ADC1->ISR & ADC_ISR_ADRDY) == 0);

	// enable temperature sensor
	ADC->CCR |= ADC_CCR_TSEN;

	ADC1->SMPR = 0; // 1.5 ADC clock cycles sampling time
	ADC1->CHSELR = ADC_CHSELR_CHSEL4; // channel 4 (motor current)

	// DMA enabled (circular), external trigger TRG1 (TIM1_CC4), rising and falling edge, overwrite old data
	ADC1->CFGR1 = ADC_CFGR1_OVRMOD | ADC_CFGR1_DMAEN | ADC_CFGR1_EXTSEL_0 | ADC_CFGR1_EXTEN;

	// clear all flags
	ADC1->ISR = 0xFF;
	ADC1->IER = 0;

	// ADC, high priority, memory increment, peripheral to memory, 16bit data size, circular mode
	DMA1_Channel1->CCR = DMA_CCR_MINC | DMA_CCR_PL_1 | DMA_CCR_MSIZE_0 | DMA_CCR_PSIZE_0 | DMA_CCR_CIRC;
	DMA1_Channel1->CMAR = (uint32_t)adc.dmaData;
	DMA1_Channel1->CPAR = (uint32_t)&ADC1->DR;
	DMA1_Channel1->CNDTR = 3;

	DMA1->IFCR = DMA_IFCR_CGIF1;

	DMA1_Channel1->CCR |= DMA_CCR_TCIE | DMA_CCR_TEIE | DMA_CCR_EN;

	NVIC_SetPriority(DMA1_Channel1_IRQn, 2);
	NVIC_EnableIRQ(DMA1_Channel1_IRQn);

	NVIC_SetPriority(ADC1_IRQn, 2);
	NVIC_EnableIRQ(ADC1_IRQn);

	ADC1->CR |= ADC_CR_ADSTART;
}

void __attribute__((optimize("O0"))) ADCWaitFor1MsTrigger()
{
	register uint32_t counter = 0;

	while(adc.trigger1ms == 0)
		++counter;

	adc.trigger1ms = 0;

	data.performance.idleCounter = counter;
	if(data.performance.idleCounter > UINT16_MAX)
		data.performance.idleCounter = UINT16_MAX;

	++data.timeMs;
}

void ADCUpdate()
{
	data.sensors.adc.vSupply_Q16_0 = adc.avgVoltage_Q16_0;
	data.sensors.adc.vSupplyReciprocal_Q0_31 = (1 << 31)/adc.avgVoltage_Q16_0;
	data.sensors.adc.iMotor_Q16_0 = adc.avgCurrent_S14_0;
}
