/*
 * ctrl.c
 *
 *  Created on: 16.03.2019
 *      Author: AndreR
 */

#include "data.h"
#include "ctrl.h"
#include "hall.h"
#include "motor.h"
#include "adc.h"

static const int16_t sinCosTable[65][2];

// hall sensor position, 60 spacing, starting at 0
static const uint16_t angleTable[] = { 0, 4096, 1365, 2731, 6827, 5461, 0, 0 };

static uint16_t getRotorAngleBlock();
static uint16_t getRotorAngleHallExtrapolation();
static uint16_t getRotorAngleEncoder();

static void dcCurrentControl(int32_t currentUVW_Q15_15[3]);
static void peakCurrentControl(int32_t currentUVW_Q15_15[3]);
static void focCurrentControl(int32_t currentUVW_Q15_15[3]);
static void testSetFixedAngle(int32_t currentUVW_Q15_15[3]);

CtrlData ctrl;

void CtrlInit()
{
	PICtrlS12Init(&data.ctrl.current, 0, 0, -512, 512);

	if(data.sensors.encoder.installed)
		ctrl.pGetRotorPosFunc = &getRotorAngleEncoder;
	else
		ctrl.pGetRotorPosFunc = &getRotorAngleHallExtrapolation;

	ctrl.pControlFunc = &testSetFixedAngle;
	ctrl.currentMode = MOTOR_MODE_ANGLE_TEST;
}

void CtrlUpdate()
{
	if(data.timeMs - data.comm.lastCommandTimeMs > 200)
	{
		data.comm.timeout = 1;
		data.comm.lastCommandTimeMs = data.timeMs - 500;
	}
	else
	{
		data.comm.timeout = 0;
	}

	if(data.comm.timeout)
	{
		data.command.motorMode = MOTOR_MODE_OFF;
		data.command.setCurrent = 0;
		data.command.setVoltage = 0;
	}

	if(data.command.motorMode == MOTOR_MODE_OFF || data.command.motorMode == MOTOR_MODE_ANGLE_TEST)
		data.command.ctrlMode = MOTOR_CTRL_NONE;

	if(data.command.flags & MOTOR_EXCHANGE_MOSI_FLAG_UPDATE_CUR_BIAS)
		data.ctrl.iMotorBias = data.sensors.adc.iMotor_Q16_0;

	data.motor.mode = data.command.motorMode;
	data.motor.angle = data.command.angle;

	data.ctrl.current.outputMax = data.sensors.adc.vSupply_Q16_0 >> 3;
	data.ctrl.current.outputMin = -data.ctrl.current.outputMax;

	PICtrlS12Enable(&data.ctrl.current, data.command.ctrlMode & MOTOR_CTRL_CURRENT);

	data.ctrl.current.kp = data.command.curCtrlKp;
	data.ctrl.current.ki = data.command.curCtrlKi;

	switch(data.command.ctrlMode)
	{
		case MOTOR_CTRL_CURRENT:
			data.ctrl.current.setpoint = data.command.setCurrent >> 1;
			break;
		default:
			data.motor.voltage = data.command.setVoltage;
			break;
	}

	if(ctrl.currentMode == data.motor.mode)
		return;

	switch(data.motor.mode)
	{
		case MOTOR_MODE_OFF: // [no break]
		case MOTOR_MODE_ANGLE_TEST:
		{
			ctrl.pControlFunc = &testSetFixedAngle;
		}
		break;
		case MOTOR_MODE_RUN:
		{
			ctrl.pControlFunc = &focCurrentControl;
		}
		break;
		case MOTOR_MODE_BLOCK:
		{
			ctrl.pControlFunc = &dcCurrentControl;
		}
		break;
	}

	ctrl.currentMode = data.motor.mode;
}

static uint16_t getRotorAngleBlock()
{
	uint16_t angle = angleTable[hall.hallPos];

	angle += 47; // 2.1 hall sensor alignment correction. TODO: individual per motor?
	angle += 455; // +20 in block mode, block is triggered 20 before hall sensor crossing // TODO: check direction sign

	return angle;
}

static uint16_t getRotorAngleHallExtrapolation()
{
	static uint8_t lastHallPos = 0;
	static int32_t comAngle_S13_8 = 0;
	static int32_t angleInc_S12_8 = 0;

	uint8_t pos = hall.hallPos;

	if(pos != lastHallPos)
	{
		// 0.09765624999999996*hallSpeed*pwmCycleSinceCom
		comAngle_S13_8 = 0;
		angleInc_S12_8 = (data.sensors.hall.speed * 25);
	}

	lastHallPos = pos;

	uint16_t angle = ((int32_t)angleTable[pos]) + (comAngle_S13_8 >> 8);
	angle += 47; // 2.1 hall sensor alignment correction. TODO: individual per motor?

	// limit angle to next hall sensor position
	if(comAngle_S13_8 < 196608 && comAngle_S13_8 > -196608)
		comAngle_S13_8 += angleInc_S12_8;

	return angle;
}

static uint16_t getRotorAngleEncoder()
{
	uint8_t pos = hall.hallPos;
	uint32_t encPosAtCom = hall.encoderPos;

	int32_t encDeltaToCom = (int32_t)motor.encDmaData[0] - (int32_t)encPosAtCom;
	int32_t exceedPos = (18432 - encDeltaToCom) >> 31;
	int32_t exceedNeg = (encDeltaToCom + 18432) >> 31;
	encDeltaToCom -= 36864 & exceedPos;
	encDeltaToCom += 36864 & exceedNeg;

	uint16_t angle = angleTable[pos] + ((encDeltaToCom * 455) >> 8);
	angle += 47; // 2.1 hall sensor alignment correction. TODO: individual per motor?
//	angle -= 455; // TODO: fix for both directions, hall sensor trigger 20 early

	return angle;
}

static void dcCurrentControl(int32_t currentUVW_Q15_15[3])
{
	static int32_t invertedFactor = 1;

	(void)currentUVW_Q15_15;

	int32_t dcCurrent_S14_0 = adc.pSampleCurrent->iMotor;

	// TODO: handle current reversal

	int32_t encDelta = (int32_t)motor.encDmaData[1] - (int32_t)motor.encDmaData[0];
	if(encDelta > 18432)
		encDelta -= 36864;
	if(encDelta < -18432)
		encDelta += 36864;

	if(encDelta < 0)
		invertedFactor = -1;
	if(encDelta > 0)
		invertedFactor = 1;

	dcCurrent_S14_0 *= invertedFactor;

	int32_t outputAdjust = 0;
	int32_t error = data.ctrl.current.setpoint - dcCurrent_S14_0;
	if(error > 0)
		outputAdjust = 4;
	if(error < 0)
		outputAdjust = -4;

	if(data.ctrl.current.enabled)
	{
		if(data.ctrl.current.setpoint == 0)
		{
			data.motor.voltage = 0;
		}
		else
		{
			data.motor.voltage += outputAdjust;
		}
	}

	if(data.motor.voltage > 65535)
		data.motor.voltage = 65535;
	if(data.motor.voltage < -65536)
		data.motor.voltage = -65536;

	data.ctrl.current.output = data.motor.voltage >> 1;

	MotorSetBlock(hall.hallPos, data.motor.voltage);
}

static void peakCurrentControl(int32_t currentUVW_Q15_15[3])
{
	// determine peak current
	int32_t signMask = currentUVW_Q15_15[0] >> 31;
	currentUVW_Q15_15[0] = (currentUVW_Q15_15[0] + signMask) ^ signMask;

	signMask = currentUVW_Q15_15[1] >> 31;
	currentUVW_Q15_15[1] = (currentUVW_Q15_15[1] + signMask) ^ signMask;

	signMask = currentUVW_Q15_15[2] >> 31;
	currentUVW_Q15_15[2] = (currentUVW_Q15_15[2] + signMask) ^ signMask;

	int32_t maxCurrent_Q15_15 = currentUVW_Q15_15[0];
	if(currentUVW_Q15_15[1] > maxCurrent_Q15_15)
		maxCurrent_Q15_15 = currentUVW_Q15_15[1];
	if(currentUVW_Q15_15[2] > maxCurrent_Q15_15)
		maxCurrent_Q15_15 = currentUVW_Q15_15[2];

	int32_t iMotor_Q14_0 = maxCurrent_Q15_15 >> 15;

	// do current control
	// TODO: handle negative current, turning direction, negative current setpoint
	int32_t outputAdjust = 0;
	int32_t error = data.ctrl.current.setpoint - iMotor_Q14_0;
	if(error > 0)
		outputAdjust = 4;
	if(error < 0)
		outputAdjust = -4;

	if(data.ctrl.current.enabled)
	{
		if(data.ctrl.current.setpoint == 0)
			data.motor.voltage = 0;
		else
			data.motor.voltage += outputAdjust;
	}

	// get current rotor angle and advance to max torque angle
	uint16_t rotorAngle = (*ctrl.pGetRotorPosFunc)();

	uint16_t const angleAdvance = 2094; // = 92

	rotorAngle += angleAdvance;

	MotorSetAngle(rotorAngle, data.motor.voltage);
}

static void focCurrentControl(int32_t currentUVW_Q15_15[3])
{
	static int32_t Uq = 0;
	static int32_t Ud = 0;

	// I_alpha = I_U
	// I_beta = 1/sqrt(3) * I_U + 2/sqrt(3) * I_V
	// (S15.15 >> 12) * S0.12 + (S15.15 >> 12) * S1.12 <=> S15.3 * S0.12 + S15.3 * S1.12 <=> S15.15 + S16.15
	int32_t currentAlpha_Q16_0 = (currentUVW_Q15_15[1] >> 15);
	int32_t currentBeta_Q16_0 = ((((currentUVW_Q15_15[1]) >> 12)*2365 + (currentUVW_Q15_15[0] >> 12)*4730) >> 15);

	uint16_t rotorAngle = (*ctrl.pGetRotorPosFunc)();
	rotorAngle &= 8191;

	int32_t sin_S0_15;
	int32_t cos_S0_15;

	CtrlSinCos_S0_15(rotorAngle, &sin_S0_15, &cos_S0_15);

	int32_t Id_Q16_0 = (cos_S0_15*currentAlpha_Q16_0 + sin_S0_15*currentBeta_Q16_0) >> 15;
	int32_t Iq_Q16_0 = (-sin_S0_15*currentAlpha_Q16_0 + cos_S0_15*currentBeta_Q16_0) >> 15;

	adc.pSampleCurrent->iMotor = Id_Q16_0;
	adc.pSampleCurrent->iMotor2 = Iq_Q16_0;

	int32_t UdAdjust = 0;
	int32_t UqAdjust = 0;

	const int32_t gainQ = 16;
	const int32_t gainD = 16;

	int32_t iDSetpoint = 0;

	int32_t IqError = data.ctrl.current.setpoint - Iq_Q16_0;
	if(IqError > 0)
		UqAdjust = gainQ;
	if(IqError < 0)
		UqAdjust = -gainQ;

	int32_t IdError = iDSetpoint - Id_Q16_0;
	if(IdError > 0)
		UdAdjust = gainD;
	if(IdError < 0)
		UdAdjust = -gainD;

	if(data.ctrl.current.enabled && data.ctrl.current.setpoint != 0)
	{
		Uq += UqAdjust;
		Ud += UdAdjust;
	}
	else
	{
		Uq = data.motor.voltage >> 3;
		Ud = 0;
	}

	// TODO: need to adjust phase voltage conversion
	if(Uq > 5190)
		Uq = 5190;
	if(Uq < -5190)
		Uq = -5190;

	if(Ud > 5190)
		Ud = 5190;
	if(Ud < -5190)
		Ud = -5190;

	data.ctrl.current.output = Ud;

	int32_t Ua_Q12_15 = cos_S0_15*Ud - sin_S0_15*Uq;
	int32_t Ub_Q12_15 = sin_S0_15*Ud + cos_S0_15*Uq;

	MotorSetAlphaBeta(Ua_Q12_15, Ub_Q12_15);
}

static void testSetFixedAngle(int32_t currentUVW_Q15_15[3])
{
	(void)currentUVW_Q15_15;

	MotorSetAngle(data.motor.angle, data.motor.voltage);
}


void CtrlUpdatePerPWMCycle(int32_t currentHighPhase_Q14_15, int32_t currentMidPhase_Q14_15)
{
//	adc.pSampleCurrent->iMotor = currentHighPhase_Q14_15 >> 15;
//	adc.pSampleCurrent->iMotor2 = currentMidPhase_Q14_15 >> 15;

	// remove offset due to PWM shifting
	currentHighPhase_Q14_15 -= 1146880;
	currentMidPhase_Q14_15 -= 4358144;

	// Iu + Iv + Iw = 0 => compute all phase currents
	int32_t currentUVW_Q15_15[3];
	currentUVW_Q15_15[motor.phIndexHigh] = currentHighPhase_Q14_15;
	currentUVW_Q15_15[motor.phIndexMid] = currentMidPhase_Q14_15 - currentHighPhase_Q14_15;
	currentUVW_Q15_15[motor.phIndexLow] = -currentMidPhase_Q14_15;

//	adc.pSampleCurrent->iMotor = currentUVW_Q15_15[0] >> 15;
//	adc.pSampleCurrent->iMotor2 = currentUVW_Q15_15[1] >> 15;

	(*ctrl.pControlFunc)(currentUVW_Q15_15);
}

// angle: 0-8191 (U13.0), sin/cos: S0.15
void CtrlSinCos_S0_15(uint16_t angle, int32_t* pSin, int32_t* pCos)
{
	uint16_t lutIndex = angle >> 7; // U6.0
	int32_t angleFractionSecond = angle & 0x7F; // S0.7
	int32_t angleFractionFirst = 127 - angleFractionSecond;

	int32_t sin_S0_22 = (int32_t)sinCosTable[lutIndex][0]*angleFractionFirst + (int32_t)sinCosTable[lutIndex+1][0]*angleFractionSecond;
	int32_t cos_S0_22 = (int32_t)sinCosTable[lutIndex][1]*angleFractionFirst + (int32_t)sinCosTable[lutIndex+1][1]*angleFractionSecond;

	*pSin = sin_S0_22 >> 7;
	*pCos = cos_S0_22 >> 7;

//	*pSin = sinCosTable[lutIndex][0];
//	*pCos = sinCosTable[lutIndex][1];
}

// format: S0.15 => 32767 = 1.0
static const int16_t sinCosTable[65][2] = {
	{      0,  32767 }, 	{   3212,  32609 }, 	{   6393,  32137 }, 	{   9512,  31356 },
	{  12539,  30273 }, 	{  15446,  28898 }, 	{  18204,  27245 }, 	{  20787,  25329 },
	{  23170,  23170 }, 	{  25329,  20787 }, 	{  27245,  18204 }, 	{  28898,  15446 },
	{  30273,  12539 }, 	{  31356,   9512 }, 	{  32137,   6393 }, 	{  32609,   3212 },
	{  32767,      0 }, 	{  32609,  -3212 }, 	{  32137,  -6393 }, 	{  31356,  -9512 },
	{  30273, -12539 }, 	{  28898, -15446 }, 	{  27245, -18204 }, 	{  25329, -20787 },
	{  23170, -23170 }, 	{  20787, -25329 }, 	{  18204, -27245 }, 	{  15446, -28898 },
	{  12539, -30273 }, 	{   9512, -31356 }, 	{   6393, -32137 }, 	{   3212, -32609 },
	{      0, -32767 }, 	{  -3212, -32609 }, 	{  -6393, -32137 }, 	{  -9512, -31356 },
	{ -12539, -30273 }, 	{ -15446, -28898 }, 	{ -18204, -27245 }, 	{ -20787, -25329 },
	{ -23170, -23170 }, 	{ -25329, -20787 }, 	{ -27245, -18204 }, 	{ -28898, -15446 },
	{ -30273, -12539 }, 	{ -31356,  -9512 }, 	{ -32137,  -6393 }, 	{ -32609,  -3212 },
	{ -32767,      0 }, 	{ -32609,   3212 }, 	{ -32137,   6393 }, 	{ -31356,   9512 },
	{ -30273,  12539 }, 	{ -28898,  15446 }, 	{ -27245,  18204 }, 	{ -25329,  20787 },
	{ -23170,  23170 }, 	{ -20787,  25329 }, 	{ -18204,  27245 }, 	{ -15446,  28898 },
	{ -12539,  30273 }, 	{  -9512,  31356 }, 	{  -6393,  32137 }, 	{  -3212,  32609 },
	{      0,  32767 }
};
