/*
 * hall.c
 *
 *  Created on: 22.01.2019
 *      Author: AndreR
 */

/**
 * Notes:
 * TIM2:
 * - Full 48MHz
 * - XOR (TI1S in CR2)
 * - Trigger selection TI1F_ED
 * - Slave in Reset Mode
 * - Timer resets with hall sensor change
 * - CCR1 in input capture on TRC => time between commutations
 * - CCR2 in PWM2 mode
 * - Triggers COM in TIM1 on CCR2 (MMS in CR2 set to OC2REF)
 * - CC1F indicates hall event and valid time measurement
 */

#include "hall.h"
#include "system_init.h"
#include "data.h"

HallData hall;

static const uint8_t reverseDir[8][8] =
{ // 	  0  1  2  3    4  5  6  7
/*0*/	{ 0, 0, 0, 0,   0, 0, 0, 0 },
/*1*/	{ 0, 0, 1, 1,   0, 0, 0, 0 },
/*2*/	{ 0, 0, 0, 0,   1, 0, 1, 0 },
/*3*/	{ 0, 0, 1, 0,   0, 0, 1, 0 },
/*4*/	{ 0, 1, 0, 0,   0, 1, 0, 0 },
/*5*/	{ 0, 1, 0, 1,   0, 0, 0, 0 },
/*6*/	{ 0, 0, 0, 0,   1, 1, 0, 0 },
/*7*/	{ 0, 0, 0, 0,   0, 0, 0, 0 },
};

void TIM2_IRQHandler()
{
	hall.encoderPos = TIM3->CNT;

	uint32_t sr = TIM2->SR;
	TIM2->SR = 0;

	uint8_t lastHall = hall.hallPos;
	hall.hallPos = HALL_GET_POS();

	if(sr & TIM_SR_CC1IF)
	{
		uint32_t cnt = TIM2->CCR1;
		hall.hallDeltaTime[0] = (int32_t)(cnt - hall.lastCCR[0]);
		hall.lastCCR[0] = cnt;
	}

	if(sr & TIM_SR_CC2IF)
	{
		uint32_t cnt = TIM2->CCR2;
		hall.hallDeltaTime[1] = (int32_t)(cnt - hall.lastCCR[1]);
		hall.lastCCR[1] = cnt;
	}

	if(sr & TIM_SR_CC3IF)
	{
		uint32_t cnt = TIM2->CCR3;
		hall.hallDeltaTime[2] = (int32_t)(cnt - hall.lastCCR[2]);
		hall.lastCCR[2] = cnt;
	}

	hall.reverseHall = reverseDir[lastHall][hall.hallPos];
}

void HallInit()
{
	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 2;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOA, GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_2, &gpioInit);

	RCC->APB1ENR |= RCC_APB1ENR_TIM2EN;

	TIM2->CR1 = 0;
	TIM2->CR2 = 0;
	TIM2->SMCR = 0;
	TIM2->SR = 0;
	TIM2->DIER = TIM_DIER_CC1IE | TIM_DIER_CC2IE | TIM_DIER_CC3IE | TIM_DIER_CC4IE;
	// IC1 is mapped to TI1, IC2 mapped to TI2, IC3 mapped to TI3
	TIM2->CCMR1 = TIM_CCMR1_CC1S_0 | (7 << TIM_CCMR1_IC1F_Pos) | TIM_CCMR1_CC2S_0 | (7 << TIM_CCMR1_IC2F_Pos);
	TIM2->CCMR2 = TIM_CCMR2_CC3S_0 | (7 << TIM_CCMR2_IC3F_Pos);
	TIM2->CCER = TIM_CCER_CC1E | TIM_CCER_CC1P | TIM_CCER_CC1NP |
			TIM_CCER_CC2E | TIM_CCER_CC2P | TIM_CCER_CC2NP | TIM_CCER_CC3E | TIM_CCER_CC3P | TIM_CCER_CC3NP;
	TIM2->CNT = 0;
	TIM2->PSC = 0;
	TIM2->ARR = UINT32_MAX;
	TIM2->EGR = TIM_EGR_UG;

	NVIC_SetPriority(TIM2_IRQn, 0);
	NVIC_EnableIRQ(TIM2_IRQn);

	TIM2->CR1 |= TIM_CR1_CEN;
}

uint32_t HallGetCommutationTime()
{
	uint16_t sr = TIM2->SR;
	uint32_t ccr1 = TIM2->CCR1;

	// did counter overflow?
	if(sr & TIM_SR_CC3IF)
	{
		if(sr & TIM_SR_UIF)
			TIM2->SR &= ~(TIM_SR_CC3IF | TIM_SR_UIF);

		// then return max value
		return UINT32_MAX;
	}

	if(sr & TIM_SR_CC1IF)
		return ccr1;

	return 0;
}

void HallSetCommutationDelay(uint32_t ticks)
{
	TIM2->CCR2 = ticks;
	TIM2->EGR = TIM_EGR_UG;
}

void HallUpdate()
{
	data.sensors.hall.position = HALL_GET_POS();

	uint32_t comTime;

	uint32_t cnt = TIM2->CNT;
	if(cnt - hall.lastCCR[0] > 16777216UL)
	{
		// no commutation, motor stopped
		hall.lastCCR[0] = cnt - 16777216UL;
		hall.lastCCR[1] = cnt - 16777216UL;
		hall.lastCCR[2] = cnt - 16777216UL;
		comTime = UINT32_MAX;
	}
	else
	{
		comTime = hall.hallDeltaTime[0] + hall.hallDeltaTime[1] + hall.hallDeltaTime[2];
		comTime = ((comTime >> 3) * 57) >> 6; // *1/9
	}

	data.sensors.hall.commutationTime = comTime;

	uint32_t speedU = (UINT32_MAX/data.sensors.hall.commutationTime) >> 7;
	data.sensors.hall.speed = (int32_t)speedU;
	if(speedU > 4095)
		speedU = 4095;

	data.sensors.hall.speedSaturated = (int32_t)speedU;

	if(hall.reverseHall)
	{
		data.sensors.hall.speed *= -1;
		data.sensors.hall.speedSaturated *= -1;
	}
}
