/*
 * main.c
 *
 *  Created on: 14.01.2019
 *      Author: AndreR
 */

/**
 * Pinout:
 *
 *  PA0: M-H1 (TIM2_CH1_ETR, AF2)
 *  PA1: M-H2 (TIM2_CH2, AF2)
 *  PA2: M-H3 (TIM2_CH3, AF2)
 *  PA3: M-VOL
 *  PA4: M-CUR
 *  PA5: -
 *  PA6: ENCA (TIM3_CH1, AF1)
 *  PA7: ENCB (TIM3_CH2, AF1)
 *  PA8: [HS1] (TIM1_CH1, AF2)
 *  PA9: [HS2] (TIM1_CH2, AF2)
 * PA10: [HS3] (TIM1_CH3, AF2)
 * PA11: [OC_SEL]
 * PA12: [OC_COMP_INT2] (TIM1_ETR, AF2)
 * PA13: SWDIO
 * PA14: SWCLK / USART1-TX
 * PA15: USART1-RX
 *
 *  PB0: -
 *  PB1: -
 *  PB3: -
 *  PB4: -
 *  PB5: -
 *  PB6: -
 *  PB7: LED
 *  PB8: -
 *  PB9: -
 * PB10: -
 * PB11: -
 * PB12: [OC_COMP_INT] (TIM1_BKIN, AF2)
 * PB13: [LS1] (TIM1_CH1N, AF2)
 * PB14: [LS2] (TIM1_CH2N, AF2)
 * PB15: [LS3] (TIM1_CH3N, AF2)
 *
 * PC13: -
 * PC14: -
 * PC15: -
 *
 *  PF0: OSC
 *  PF1: -
 *  PF6: [OC_TH_STBY2]
 *  PF7: [OC_TH_STBY1]
 */

// motor speed control
// Input type: encoder delta position within 1ms
// Input range: 0 - 6144 (0 - 10,000RPM) Q14.0
// Minimum resolution: approx. 1mm/s at wheel
// Output type: PWM value
// Output range: 0 - 4096 Q12.0
// PI controller gains: Ki Q0.17 (max. 1, eps: ~7.6e-6), Kp Q7.10 (max. 128, eps: ~0.001)
// Output limits: outMin, outMax

#include "system_init.h"
#include "usart1.h"
#include "enc.h"
#include "hall.h"
#include "motor.h"
#include "adc.h"
#include "ctrl.h"
#include "log_msgs.h"
#include "pi_ctrl_s12.h"
#include "data.h"

#if ADC_NUM_SAMPLES_PER_MS != MOTOR_EXCHANGE_MISO_MAX_MEAS
#error Number of ADC Samples and motor exchange measurements must match
#endif

#define SYSTEM_CLOCK_HZ 48000000UL
#define SYSTEM_TICK_FREQ 1000

#define TOGGLE_PRIMARY_LED() GPIOB->ODR ^= GPIO_ODR_7

Data data;

int main()
{
	SystemInit(SYSTEM_CLOCK_HZ, SYSTEM_TICK_FREQ);

	RCC->AHBENR |= RCC_AHBENR_GPIOAEN | RCC_AHBENR_GPIOBEN | RCC_AHBENR_GPIOFEN | RCC_AHBENR_DMAEN;

	// primary LED
	GPIOB->MODER |= GPIO_MODER_MODER7_0;
	GPIOB->BSRR |= GPIO_BSRR_BR_7;

	GPIOInitData gpio;
	gpio.mode = GPIO_MODE_OUTPUT;
	gpio.otype = GPIO_OTYPE_PUSH_PULL;
	gpio.ospeed = GPIO_OSPEED_50MHZ;
	gpio.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOA, GPIO_PIN_13, &gpio);

	GPIOReset(GPIOA, GPIO_PIN_13);

	data.performance.idleIncrements = SystemMeasureIdleIncrements();
	if(data.performance.idleIncrements > UINT16_MAX)
		data.performance.idleIncrements = UINT16_MAX;

	EncInit();
	USART1Init();
	HallInit();
	ADCInit();
	MotorInit();
	CtrlInit();

	// configure independent watchdog
	IWDG->KR = 0xCCCC;
	IWDG->KR = 0x5555;
	IWDG->PR = 0; // div4 => 10kHz
	IWDG->RLR = 100; // => 10ms timeout
	while(IWDG->SR != 0);
	IWDG->KR = 0xAAAA;

	while(1)
	{
		ADCWaitFor1MsTrigger();

		IWDG->KR = 0xAAAA; // reload IWDG counter value to avoid reset
		GPIOA->ODR ^= GPIO_PIN_13;

		EncUpdate();
		HallUpdate();
		ADCUpdate();

		CtrlUpdate();

		MotorUpdate();

		USART1SendFeedback();
		USART1ReceiveCommand();

		if(data.timeMs % 512 == 0)
			TOGGLE_PRIMARY_LED();
	}

	return 0;
}
