/*
 * motor.c
 *
 *  Created on: 25.01.2019
 *      Author: AndreR
 */

#include "motor.h"
#include "system_init.h"
#include "hall.h"
#include "log_msgs.h"
#include "data.h"
#include "ctrl.h"

/**
 * Motor notes:
 * - CCR4 used for ADC trigger => possible to start before center of PWM
 * - Phase off: CCxE=1, CCxNE=0, force inactive
 * - Phase low: CCxE=0, CCxNE=1, force active
 * - Phase PWM: CCxE=1, CCxNE=1, PWM mode 1
 */

#define CCMR1_PHASE1_OFF (TIM_CCMR1_OC1PE | TIM_CCMR1_OC1M_2)
#define  CCER_PHASE1_OFF (TIM_CCER_CC1E)
#define CCMR1_PHASE1_LOW (TIM_CCMR1_OC1PE | TIM_CCMR1_OC1M_0 | TIM_CCMR1_OC1M_2)
#define  CCER_PHASE1_LOW (TIM_CCER_CC1NE)
#define CCMR1_PHASE1_HIGH (TIM_CCMR1_OC1PE | TIM_CCMR1_OC1M_0 | TIM_CCMR1_OC1M_2)
#define  CCER_PHASE1_HIGH (TIM_CCER_CC1E)
#define CCMR1_PHASE1_PWM (TIM_CCMR1_OC1PE | TIM_CCMR1_OC1M_1 | TIM_CCMR1_OC1M_2 | TIM_CCMR1_OC1CE)
#define  CCER_PHASE1_PWM (TIM_CCER_CC1E | TIM_CCER_CC1NE)
#define CCMR1_PHASE1_PWM2 (TIM_CCMR1_OC1PE | TIM_CCMR1_OC1M_0 | TIM_CCMR1_OC1M_1 | TIM_CCMR1_OC1M_2 | TIM_CCMR1_OC1CE)
#define  CCER_PHASE1_PWM2 (TIM_CCER_CC1E | TIM_CCER_CC1NE)

#define CCMR1_PHASE2_OFF (TIM_CCMR1_OC2PE | TIM_CCMR1_OC2M_2)
#define  CCER_PHASE2_OFF (TIM_CCER_CC2E)
#define CCMR1_PHASE2_LOW (TIM_CCMR1_OC2PE | TIM_CCMR1_OC2M_0 | TIM_CCMR1_OC2M_2)
#define  CCER_PHASE2_LOW (TIM_CCER_CC2NE)
#define CCMR1_PHASE2_HIGH (TIM_CCMR1_OC2PE | TIM_CCMR1_OC2M_0 | TIM_CCMR1_OC2M_2)
#define  CCER_PHASE2_HIGH (TIM_CCER_CC2E)
#define CCMR1_PHASE2_PWM (TIM_CCMR1_OC2PE | TIM_CCMR1_OC2M_1 | TIM_CCMR1_OC2M_2 | TIM_CCMR1_OC2CE)
#define  CCER_PHASE2_PWM (TIM_CCER_CC2E | TIM_CCER_CC2NE)
#define CCMR1_PHASE2_PWM2 (TIM_CCMR1_OC2PE | TIM_CCMR1_OC2M_0 | TIM_CCMR1_OC2M_1 | TIM_CCMR1_OC2M_2 | TIM_CCMR1_OC2CE)
#define  CCER_PHASE2_PWM2 (TIM_CCER_CC2E | TIM_CCER_CC2NE)

#define CCMR2_PHASE3_OFF (TIM_CCMR2_OC3PE | TIM_CCMR2_OC3M_2)
#define  CCER_PHASE3_OFF (TIM_CCER_CC3E)
#define CCMR2_PHASE3_LOW (TIM_CCMR2_OC3PE | TIM_CCMR2_OC3M_0 | TIM_CCMR2_OC3M_2)
#define  CCER_PHASE3_LOW (TIM_CCER_CC3NE)
#define CCMR2_PHASE3_HIGH (TIM_CCMR2_OC3PE | TIM_CCMR2_OC3M_0 | TIM_CCMR2_OC3M_2)
#define  CCER_PHASE3_HIGH (TIM_CCER_CC3E)
#define CCMR2_PHASE3_PWM (TIM_CCMR2_OC3PE | TIM_CCMR2_OC3M_1 | TIM_CCMR2_OC3M_2 | TIM_CCMR2_OC3CE)
#define  CCER_PHASE3_PWM (TIM_CCER_CC3E | TIM_CCER_CC3NE)
#define CCMR2_PHASE3_PWM2 (TIM_CCMR2_OC3PE | TIM_CCMR2_OC3M_0 | TIM_CCMR2_OC3M_1 | TIM_CCMR2_OC3M_2 | TIM_CCMR2_OC3CE)
#define  CCER_PHASE3_PWM2 (TIM_CCER_CC3E | TIM_CCER_CC3NE)

// Channel 4 is used to trigger ADC in block commutation mode, PWM mode 1
#define CCMR2_PHASE4_PWM (TIM_CCMR2_OC4M_1 | TIM_CCMR2_OC4M_2)
#define  CCER_PHASE4_PWM (TIM_CCER_CC4E)

// Channel 4 is used to trigger ADC in sine commutation mode, toggle mode
#define CCMR2_PHASE4_TOG (TIM_CCMR2_OC4M_0 | TIM_CCMR2_OC4M_1)
#define  CCER_PHASE4_TOG (TIM_CCER_CC4E)

typedef struct _ComEntry
{
	uint16_t ccmr1;
	uint16_t ccmr2;
	uint16_t ccer;
} ComEntry;

static const ComEntry comTable[] = {
	{ // invalid hall position => all phases off
		CCMR1_PHASE1_OFF | CCMR1_PHASE2_OFF, CCMR2_PHASE3_OFF | CCMR2_PHASE4_TOG,
		 CCER_PHASE1_OFF |  CCER_PHASE2_OFF | CCER_PHASE3_OFF |  CCER_PHASE4_TOG
	},
	{ // hall pos 1 => 1 low, 2 off, 3 PWM
		CCMR1_PHASE1_LOW | CCMR1_PHASE2_OFF, CCMR2_PHASE3_PWM | CCMR2_PHASE4_TOG,
		 CCER_PHASE1_LOW |  CCER_PHASE2_OFF | CCER_PHASE3_PWM |  CCER_PHASE4_TOG
	},
	{ // hall pos 2 => 1 PWM, 2 low, 3 off
		CCMR1_PHASE1_PWM | CCMR1_PHASE2_LOW, CCMR2_PHASE3_OFF | CCMR2_PHASE4_TOG,
		 CCER_PHASE1_PWM |  CCER_PHASE2_LOW | CCER_PHASE3_OFF |  CCER_PHASE4_TOG
	},
	{ // hall pos 3 => 1 off, 2 low, 3 PWM
		CCMR1_PHASE1_OFF | CCMR1_PHASE2_LOW, CCMR2_PHASE3_PWM | CCMR2_PHASE4_TOG,
		 CCER_PHASE1_OFF |  CCER_PHASE2_LOW | CCER_PHASE3_PWM |  CCER_PHASE4_TOG
	},
	{ // hall pos 4 => 1 off, 2 PWM, 3 low
		CCMR1_PHASE1_OFF | CCMR1_PHASE2_PWM, CCMR2_PHASE3_LOW | CCMR2_PHASE4_TOG,
		 CCER_PHASE1_OFF |  CCER_PHASE2_PWM | CCER_PHASE3_LOW |  CCER_PHASE4_TOG
	},
	{ // hall pos 5 => 1 low, 2 PWM, 3 off
		CCMR1_PHASE1_LOW | CCMR1_PHASE2_PWM, CCMR2_PHASE3_OFF | CCMR2_PHASE4_TOG,
		 CCER_PHASE1_LOW |  CCER_PHASE2_PWM | CCER_PHASE3_OFF |  CCER_PHASE4_TOG
	},
	{ // hall pos 6 => 1 PWM, 2 off, 3 low
		CCMR1_PHASE1_PWM | CCMR1_PHASE2_OFF, CCMR2_PHASE3_LOW | CCMR2_PHASE4_TOG,
		 CCER_PHASE1_PWM |  CCER_PHASE2_OFF | CCER_PHASE3_LOW |  CCER_PHASE4_TOG
	},
	{ // invalid hall position => all low sides on
		CCMR1_PHASE1_LOW | CCMR1_PHASE2_LOW, CCMR2_PHASE3_LOW | CCMR2_PHASE4_TOG,
		 CCER_PHASE1_LOW |  CCER_PHASE2_LOW | CCER_PHASE3_LOW |  CCER_PHASE4_TOG
	}
};

static const AngleLUTEntry angleLUT[512];

MotorData motor;

void TIM1_BRK_UP_TRG_COM_IRQHandler()
{
	uint32_t sr = TIM1->SR;
	TIM1->SR &= ~(TIM_SR_BIF | TIM_SR_UIF);

	if(sr & TIM_SR_BIF)
	{
		data.motor.breakActive = 1;

		TIM1->CCMR1 = CCMR1_PHASE1_OFF | CCMR1_PHASE2_OFF;
		TIM1->CCMR2 = CCMR2_PHASE3_OFF | CCMR2_PHASE4_TOG;
		TIM1->CCER = CCER_PHASE1_OFF | CCER_PHASE2_OFF | CCER_PHASE3_OFF | CCER_PHASE4_TOG;
		TIM1->EGR = TIM_EGR_UG | TIM_EGR_COMG;
		TIM1->BDTR |= TIM_BDTR_MOE;
	}
}

void MotorInit()
{
	motor.currentMode = MOTOR_MODE_OFF;

	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 2;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_2MHZ;
	gpioInit.pupd = GPIO_PUPD_DOWN;
	GPIOInit(GPIOA, GPIO_PIN_8 | GPIO_PIN_9 | GPIO_PIN_10, &gpioInit); // HS1, HS2, HS3
	GPIOInit(GPIOA, GPIO_PIN_12, &gpioInit); // OC_COMP => TIM1_ETR
	GPIOInit(GPIOB, GPIO_PIN_13 | GPIO_PIN_14 | GPIO_PIN_15, &gpioInit); // LS1, LS2, LS3
	GPIOInit(GPIOB, GPIO_PIN_12, &gpioInit); // OC_COMP => TIM1_BKIN

	gpioInit.mode = GPIO_MODE_OUTPUT;
	GPIOInit(GPIOA, GPIO_PIN_11, &gpioInit); // OC_SEL
	GPIOInit(GPIOF, GPIO_PIN_6 | GPIO_PIN_7, &gpioInit); // OC_TH_STBY2, OC_TH_STBY1

	// OC_SEL = 0 => OC comparator output signal is visible only to MCU (default)
	GPIOReset(GPIOA, GPIO_PIN_11);

	// Set OC threshold to 500mV
	GPIOSet(GPIOF, GPIO_PIN_6 | GPIO_PIN_7);

	RCC->APB2ENR |= RCC_APB2ENR_TIM1EN;

	TIM1->PSC = 0; // => 48MHz
	TIM1->ARR = 1199; // => 40kHz => 20kHz due to center-aligned mode
	// ARR is preloaded, Center-Aligned Mode 3
	TIM1->CR1 = TIM_CR1_ARPE | TIM_CR1_CMS;
	// CCxE, CCxNE, OCxM are preloaded, update with COM
	TIM1->CR2 = TIM_CR2_CCPC;
//	TIM1->SMCR = 0;
	TIM1->SMCR = TIM_SMCR_OCCS | TIM_SMCR_ETF_3;
	TIM1->DIER = 0;
	TIM1->SR = 0;
	// CCR registers preloaded, all half-bridges off, OCxRef cleared by ETRF
	TIM1->CCMR1 = CCMR1_PHASE1_OFF | CCMR1_PHASE2_OFF;
	TIM1->CCMR2 = CCMR2_PHASE3_OFF | CCMR2_PHASE4_TOG;
	TIM1->CCER = CCER_PHASE1_OFF | CCER_PHASE2_OFF | CCER_PHASE3_OFF | CCER_PHASE4_TOG;
	TIM1->CCR1 = 0;
	TIM1->CCR2 = 0;
	TIM1->CCR3 = 0;
	TIM1->CCR4 = 1056; // 3us before counter overflow
	TIM1->EGR = TIM_EGR_UG | TIM_EGR_COMG;
	// Outputs forced to inactive level in idle (MOE=0), 187.5ns dead time, active high break input
//	TIM1->BDTR = TIM_BDTR_OSSR | TIM_BDTR_OSSI | (9 << TIM_BDTR_DTG_Pos) | TIM_BDTR_BKE | TIM_BDTR_BKP;
	TIM1->BDTR = TIM_BDTR_OSSR | TIM_BDTR_OSSI | (9 << TIM_BDTR_DTG_Pos) | TIM_BDTR_BKP;
	TIM1->DCR = (3 << TIM_DCR_DBL_Pos) | (13 << TIM_DCR_DBA_Pos);

	NVIC_SetPriority(TIM1_BRK_UP_TRG_COM_IRQn, 1);
	NVIC_EnableIRQ(TIM1_BRK_UP_TRG_COM_IRQn);

	// DMA1_CH4 used to write new timer data to CCMR registers
	// TIM1_CH4, very high priority, memory increment, memory to peripheral, 16bit data size, circular mode
	DMA1_Channel4->CCR = DMA_CCR_MINC | DMA_CCR_PL | DMA_CCR_MSIZE_0 | DMA_CCR_PSIZE_0 | DMA_CCR_CIRC | DMA_CCR_DIR;
	DMA1_Channel4->CMAR = (uint32_t)motor.dmaData;
	DMA1_Channel4->CPAR = (uint32_t)&TIM1->DMAR;
	DMA1_Channel4->CNDTR = 12;

	DMA1->IFCR = DMA_IFCR_CGIF4;
	DMA1_Channel4->CCR |= DMA_CCR_EN;

	// DMA1_CH5 used to capture encoder position
	// TIM1_UP, high priority, memory increment, peripheral to memory, 16bit data size, circular mode
	DMA1_Channel5->CCR = DMA_CCR_MINC | DMA_CCR_PL_1 | DMA_CCR_MSIZE_0 | DMA_CCR_PSIZE_0 | DMA_CCR_CIRC;
	DMA1_Channel5->CMAR = (uint32_t)motor.encDmaData;
	DMA1_Channel5->CPAR = (uint32_t)&TIM3->CNT;
	DMA1_Channel5->CNDTR = 2;

	DMA1->IFCR = DMA_IFCR_CGIF5;
	DMA1_Channel5->CCR |= DMA_CCR_EN;

	motor.dmaData[0]  = 600;
	motor.dmaData[1]  = 600;
	motor.dmaData[2]  = 600;
	motor.dmaData[3]  = 700;
	motor.dmaData[4]  = 600;
	motor.dmaData[5]  = 600;
	motor.dmaData[6]  = 600;
	motor.dmaData[7]  = 500;
	motor.dmaData[8]  = 600;
	motor.dmaData[9]  = 600;
	motor.dmaData[10] = 600;
	motor.dmaData[11] = 1128;
	motor.phIndexLow = 0;
	motor.phIndexMid = 1;
	motor.phIndexHigh = 2;

	TIM1->CR1 |= TIM_CR1_CEN;
	TIM1->BDTR |= TIM_BDTR_MOE;
	TIM1->DIER |= TIM_DIER_BIE | TIM_DIER_UDE | TIM_DIER_CC4DE;
}

static inline int16_t voltageToPwm(int32_t voltage)
{
	int32_t pwm; // Q12.0

	if(voltage >= data.sensors.adc.vSupply_Q16_0)
		pwm = 4095;
	else if(voltage <= -data.sensors.adc.vSupply_Q16_0)
		pwm = -4096;
	else
		pwm = ((voltage * data.sensors.adc.vSupplyReciprocal_Q0_31) >> 19);

	return pwm;
}

static inline void getPhaseVoltages(int32_t alpha_S12_15, int32_t beta_S12_15, int32_t* pVp)
{
	// Vb = 1/sqrt(3) * Valpha;
	// Va = 1/(sqrt(3)*2) * Valpha + 0.5*Vbeta;
	// Vc = 1/(sqrt(3)*2) * Valpha - 0.5*Vbeta;

	int32_t alphaSqrt3_S12_15 = ((alpha_S12_15 >> 6)*591) >> 4;
	int32_t alphaSqrt3HalfNeg_S12_15 = -(alphaSqrt3_S12_15 >> 1);
	int32_t betaHalf_S12_15 = beta_S12_15 >> 1;

	pVp[1] = alphaSqrt3_S12_15;
	pVp[0] = alphaSqrt3HalfNeg_S12_15 + betaHalf_S12_15;
	pVp[2] = alphaSqrt3HalfNeg_S12_15 - betaHalf_S12_15;
}

// alpha, beta: +-4096 PWM value
static void setAlphaBeta7Segment(int32_t alpha_S12_15, int32_t beta_S12_15)
{
	int32_t Vp[3];

	getPhaseVoltages(alpha_S12_15, beta_S12_15, Vp);
	
	// convert +-4096 to +-600 (CCR range)
	Vp[0] = ((Vp[0] >> 5) * 75) >> 4;
	Vp[1] = ((Vp[1] >> 5) * 75) >> 4;
	Vp[2] = ((Vp[2] >> 5) * 75) >> 4;

	// compare and LUT to find min, mid, max phase voltages
	// Sector:                            0  1  2  3  4  5
	uint8_t aLargerB = Vp[0] >= Vp[1]; // 0  1  1  1  0  0
	uint8_t bLargerC = Vp[1] >= Vp[2]; // 1  1  0  0  0  1
	uint8_t cLargerA = Vp[2] > Vp[0];  // 0  0  0  1  1  1
	uint8_t sectorGray = aLargerB | bLargerC << 1 | cLargerA << 2;

	static const uint8_t vLowIndex[] = { 0, 1, 2, 2, 0, 1, 0, 0 }; // -i
	static const uint8_t vMidIndex[] = { 0, 2, 0, 1, 1, 0, 2, 0 };
	static const uint8_t vHigIndex[] = { 0, 0, 1, 0, 2, 2, 1, 0 }; // +i

	motor.phIndexLow = vLowIndex[sectorGray];
	motor.phIndexMid = vMidIndex[sectorGray];
	motor.phIndexHigh = vHigIndex[sectorGray];

	int32_t VLow = Vp[motor.phIndexLow];
	int32_t VHigh = Vp[motor.phIndexHigh];

	// Offset voltages to center at 50%, otherwise they could be negative
	int32_t Vcom = (VLow+VHigh) >> 1;
	Vp[0] = ((Vp[0]-Vcom) >> 15) + 600;
	Vp[1] = ((Vp[1]-Vcom) >> 15) + 600;
	Vp[2] = ((Vp[2]-Vcom) >> 15) + 600;

	// Update low, mid, high phase voltages after offset shift
	VLow = Vp[motor.phIndexLow];
	int32_t VMid = Vp[motor.phIndexMid];
	VHigh = Vp[motor.phIndexHigh];

	motor.VHighUsed_U12_0 = (VHigh * 13981) >> 12;
	motor.VMidUsed_U12_0 = (VMid * 13981) >> 12;
	motor.VLowUsed_U12_0 = (VLow * 13981) >> 12;

	// when to turn PWM off in second half of center-aligned mode
	int32_t VLowOff = VLow;
	int32_t VMidOff = VMid;
	int32_t VHighOff = VHigh;

	// required minimum time for ADC sampling
	const int32_t tCrit = 79; // = 1.646us

	// adjust high and mid phase voltages to measure current
	int32_t VMidLow = VMid - VLow;
	int32_t midLowSmallDelta = (VMidLow - tCrit) >> 31;
	VMid += (tCrit-VMidLow) & midLowSmallDelta;
	VMidOff -= (tCrit-VMidLow) & midLowSmallDelta;

	int32_t VHighMid = VHigh - VMid;
	int32_t highMidSmallDelta = (VHighMid - tCrit) >> 31;
	VHigh += (tCrit-VHighMid) & highMidSmallDelta;
	VHighOff -= (tCrit-VHighMid) & highMidSmallDelta;

	// ADC trigger points
	int32_t adcHigh = ((VHigh+VMid) >> 1) + 0;
	int32_t adcMid = ((VMid+VLow) >> 1) + 0;

	// set on times
	motor.dmaData[motor.phIndexLow] = VLow;
	motor.dmaData[motor.phIndexMid] = VMid;
	motor.dmaData[motor.phIndexHigh] = VHigh;
//	// set ADC trigger points
	motor.dmaData[3] = adcHigh;
	motor.dmaData[7] = adcMid;
//	// set off times
	motor.dmaData[8+motor.phIndexLow] = VLowOff;
	motor.dmaData[8+motor.phIndexMid] = VMidOff;
	motor.dmaData[8+motor.phIndexHigh] = VHighOff;
}

// angle: 0-8191, voltage: +-65536 [mV]
static inline void setAngle(uint16_t angle, int32_t voltage)
{
	int32_t sin_S0_15;
	int32_t cos_S0_15;

	angle &= 8191;

	CtrlSinCos_S0_15(angle, &sin_S0_15, &cos_S0_15);

	int32_t pwm_S12_0 = voltageToPwm(voltage);

	int32_t alpha_S12_15 = cos_S0_15*pwm_S12_0;
	int32_t beta_S12_15 = sin_S0_15*pwm_S12_0;

	setAlphaBeta7Segment(alpha_S12_15, beta_S12_15);
}

// angle: 0-8191, voltage: +-65536 [mV]
static inline void setAngleLUT(uint16_t angle, int32_t voltage)
{
	// absolute value without branching
	int32_t signMask = voltage >> 31; // all ones if voltage is negative, all zeros otherwise
	voltage = (voltage + signMask) ^ signMask; // voltage is now always positive

	// rotate angle by 180 if voltage is negative
	angle += (4096 & signMask);
	angle &= 8191;

	// clamp voltage first
	int32_t exceededMask = (data.sensors.adc.vSupply_Q16_0 - voltage) >> 31; // all ones if supply voltage exceeded

	int32_t pwm_S0_12 = ((voltage * data.sensors.adc.vSupplyReciprocal_Q0_31) >> 19);

	// positive clamping
	pwm_S0_12 |= exceededMask;
	pwm_S0_12 &= 0x0FFF;

	const AngleLUTEntry* pEntry = &angleLUT[angle >> 4];

	const int32_t VOffset_10_17 = 78643200; // 600 * 2^17

	int32_t VHigh_U11_0 = (pEntry->VHigh_S10_5 * pwm_S0_12 + VOffset_10_17) >> 17;
	int32_t VMid_U11_0 = (pEntry->VMid_S10_5 * pwm_S0_12 + VOffset_10_17) >> 17;
	int32_t VLow_U11_0 = (pEntry->VLow_S10_5 * pwm_S0_12 + VOffset_10_17) >> 17;

//	motor.VHighUsed = pEntry->VHigh_S10_5;
//	motor.VMidUsed = pEntry->VMid_S10_5;
//	motor.VLowUsed = pEntry->VLow_S10_5;
	motor.VHighUsed_U12_0 = (VHigh_U11_0 * 13981) >> 12;
	motor.VMidUsed_U12_0 = (VMid_U11_0 * 13981) >> 12;
	motor.VLowUsed_U12_0 = (VLow_U11_0 * 13981) >> 12;
	motor.phIndexLow = pEntry->lowIndex;
	motor.phIndexMid = pEntry->midIndex;
	motor.phIndexHigh = pEntry->highIndex;

	// when to turn PWM off in second half of center-aligned mode
	int32_t VLowOff = VLow_U11_0;
	int32_t VMidOff = VMid_U11_0;
	int32_t VHighOff = VHigh_U11_0;

	// required minimum time for ADC sampling
	const int32_t tCrit = 79; // = 1.646us

	int32_t VMidLow = VMid_U11_0 - VLow_U11_0;
	int32_t midLowSmallDelta = (VMidLow - tCrit) >> 31;
	VMid_U11_0 += (tCrit-VMidLow) & midLowSmallDelta;
	VMidOff -= (tCrit-VMidLow) & midLowSmallDelta;

	int32_t VHighMid = VHigh_U11_0 - VMid_U11_0;
	int32_t highMidSmallDelta = (VHighMid - tCrit) >> 31;
	VHigh_U11_0 += (tCrit-VHighMid) & highMidSmallDelta;
	VHighOff -= (tCrit-VHighMid) & highMidSmallDelta;

	// ADC trigger points
	int32_t adcHigh = ((VHigh_U11_0+VMid_U11_0) >> 1) + 0;
	int32_t adcMid = ((VMid_U11_0+VLow_U11_0) >> 1) + 0;

	// set on times
	motor.dmaData[pEntry->lowIndex] = VLow_U11_0;
	motor.dmaData[pEntry->midIndex] = VMid_U11_0;
	motor.dmaData[pEntry->highIndex] = VHigh_U11_0;
//	// set ADC trigger points
	motor.dmaData[3] = adcHigh;
	motor.dmaData[7] = adcMid;
//	// set off times
	motor.dmaData[8+pEntry->lowIndex] = VLowOff;
	motor.dmaData[8+pEntry->midIndex] = VMidOff;
	motor.dmaData[8+pEntry->highIndex] = VHighOff;
}

// angle: 0-8191, voltage: +-65536 [mV]
void MotorSetAngle(uint16_t angle, int32_t voltage)
{
	setAngleLUT(angle, voltage);
}

// alpha, beta: +-4096 PWM value
void MotorSetAlphaBeta(int32_t alpha_S12_15, int32_t beta_S12_15)
{
	setAlphaBeta7Segment(alpha_S12_15, beta_S12_15);
}

// voltage: +-65536 [mV]
void MotorSetBlock(uint8_t phase, int32_t voltage)
{
	if(voltage < 0)
	{
		phase ^= 0x07;
		voltage *= -1;
	}

	int32_t pwm_S12_0 = voltageToPwm(voltage);

	motor.VHighUsed_U12_0 = pwm_S12_0;
	motor.VMidUsed_U12_0 = pwm_S12_0;
	motor.VLowUsed_U12_0 = 0;

	// convert 0-4096 to 0-1200 (CCR range)
	int32_t ccr = (pwm_S12_0 * 1200) >> 12;

	motor.dmaData[0]  = ccr;
	motor.dmaData[1]  = ccr;
	motor.dmaData[2]  = ccr;
	motor.dmaData[3]  = 600;

	// minimum time required to load new CCR values via DMA
	motor.dmaData[7]  = 38;

	motor.dmaData[8]  = ccr;
	motor.dmaData[9]  = ccr;
	motor.dmaData[10] = ccr;
	motor.dmaData[11] = 1128;

	const ComEntry* pCom = &comTable[phase];

	TIM1->CCMR1 = pCom->ccmr1;
	TIM1->CCMR2 = pCom->ccmr2;
	TIM1->CCER = pCom->ccer;
	TIM1->EGR = TIM_EGR_COMG;
}

void MotorUpdate()
{
	if(data.motor.breakActive && (data.command.flags & MOTOR_EXCHANGE_MOSI_FLAG_CLEAR_BREAK) && data.sensors.adc.iMotor_Q16_0 < 1000)
		data.motor.breakActive = 0;

	if(data.motor.breakActive)
		return;

	if(motor.currentMode == data.motor.mode)
		return;

	switch(data.motor.mode)
	{
		case MOTOR_MODE_OFF:
		{
			TIM1->CCMR1 = CCMR1_PHASE1_OFF | CCMR1_PHASE2_OFF;
			TIM1->CCMR2 = CCMR2_PHASE3_OFF | CCMR2_PHASE4_TOG;
			TIM1->CCER = CCER_PHASE1_OFF | CCER_PHASE2_OFF | CCER_PHASE3_OFF | CCER_PHASE4_TOG;
			TIM1->EGR = TIM_EGR_COMG;
		}
		break;
		case MOTOR_MODE_ANGLE_TEST: // [no break]
		case MOTOR_MODE_RUN:
		{
			TIM1->CCMR1 = CCMR1_PHASE1_PWM | CCMR1_PHASE2_PWM;
			TIM1->CCMR2 = CCMR2_PHASE3_PWM | CCMR2_PHASE4_TOG;
			TIM1->CCER = CCER_PHASE1_PWM | CCER_PHASE2_PWM | CCER_PHASE3_PWM |  CCER_PHASE4_TOG;
			TIM1->EGR = TIM_EGR_COMG;

			TIM2->EGR = TIM_EGR_CC4G;
		}
		break;
		case MOTOR_MODE_BLOCK:
		{
			// CCMR, CCER configuration is updated on next commutation

			// trigger hall sensor update
			TIM2->EGR = TIM_EGR_CC4G;
		}
		break;
	}

	motor.currentMode = data.motor.mode;
}

static const AngleLUTEntry angleLUT[] =
{
	{ -16629, -16629,  16629,  0,  2,  1 }, 	{ -16746, -16274,  16746,  2,  0,  1 },
	{ -16860, -15917,  16860,  2,  0,  1 }, 	{ -16971, -15558,  16971,  2,  0,  1 },
	{ -17080, -15196,  17080,  2,  0,  1 }, 	{ -17187, -14832,  17187,  2,  0,  1 },
	{ -17290, -14465,  17290,  2,  0,  1 }, 	{ -17391, -14097,  17391,  2,  0,  1 },
	{ -17490, -13726,  17490,  2,  0,  1 }, 	{ -17586, -13353,  17586,  2,  0,  1 },
	{ -17679, -12979,  17679,  2,  0,  1 }, 	{ -17770, -12602,  17770,  2,  0,  1 },
	{ -17858, -12223,  17858,  2,  0,  1 }, 	{ -17943, -11843,  17943,  2,  0,  1 },
	{ -18026, -11461,  18026,  2,  0,  1 }, 	{ -18105, -11077,  18105,  2,  0,  1 },
	{ -18182, -10691,  18182,  2,  0,  1 }, 	{ -18257, -10304,  18257,  2,  0,  1 },
	{ -18328,  -9915,  18328,  2,  0,  1 }, 	{ -18397,  -9525,  18397,  2,  0,  1 },
	{ -18463,  -9133,  18463,  2,  0,  1 }, 	{ -18527,  -8740,  18527,  2,  0,  1 },
	{ -18587,  -8345,  18587,  2,  0,  1 }, 	{ -18645,  -7950,  18645,  2,  0,  1 },
	{ -18700,  -7553,  18700,  2,  0,  1 }, 	{ -18752,  -7155,  18752,  2,  0,  1 },
	{ -18801,  -6756,  18801,  2,  0,  1 }, 	{ -18848,  -6356,  18848,  2,  0,  1 },
	{ -18891,  -5955,  18891,  2,  0,  1 }, 	{ -18932,  -5553,  18932,  2,  0,  1 },
	{ -18970,  -5150,  18970,  2,  0,  1 }, 	{ -19005,  -4746,  19005,  2,  0,  1 },
	{ -19037,  -4342,  19037,  2,  0,  1 }, 	{ -19066,  -3937,  19066,  2,  0,  1 },
	{ -19093,  -3532,  19093,  2,  0,  1 }, 	{ -19116,  -3125,  19116,  2,  0,  1 },
	{ -19137,  -2719,  19137,  2,  0,  1 }, 	{ -19155,  -2312,  19155,  2,  0,  1 },
	{ -19170,  -1905,  19170,  2,  0,  1 }, 	{ -19182,  -1497,  19182,  2,  0,  1 },
	{ -19191,  -1089,  19191,  2,  0,  1 }, 	{ -19197,   -681,  19197,  2,  0,  1 },
	{ -19201,   -273,  19201,  2,  0,  1 }, 	{ -19201,    135,  19201,  2,  0,  1 },
	{ -19199,    543,  19199,  2,  0,  1 }, 	{ -19193,    951,  19193,  2,  0,  1 },
	{ -19185,   1359,  19185,  2,  0,  1 }, 	{ -19174,   1766,  19174,  2,  0,  1 },
	{ -19160,   2174,  19160,  2,  0,  1 }, 	{ -19143,   2581,  19143,  2,  0,  1 },
	{ -19123,   2988,  19123,  2,  0,  1 }, 	{ -19101,   3394,  19101,  2,  0,  1 },
	{ -19075,   3800,  19075,  2,  0,  1 }, 	{ -19047,   4205,  19047,  2,  0,  1 },
	{ -19016,   4609,  19016,  2,  0,  1 }, 	{ -18982,   5013,  18982,  2,  0,  1 },
	{ -18945,   5416,  18945,  2,  0,  1 }, 	{ -18905,   5818,  18905,  2,  0,  1 },
	{ -18862,   6220,  18862,  2,  0,  1 }, 	{ -18817,   6620,  18817,  2,  0,  1 },
	{ -18768,   7020,  18768,  2,  0,  1 }, 	{ -18717,   7418,  18717,  2,  0,  1 },
	{ -18663,   7815,  18663,  2,  0,  1 }, 	{ -18606,   8211,  18606,  2,  0,  1 },
	{ -18547,   8606,  18547,  2,  0,  1 }, 	{ -18484,   9000,  18484,  2,  0,  1 },
	{ -18419,   9392,  18419,  2,  0,  1 }, 	{ -18351,   9783,  18351,  2,  0,  1 },
	{ -18281,  10172,  18281,  2,  0,  1 }, 	{ -18207,  10560,  18207,  2,  0,  1 },
	{ -18131,  10946,  18131,  2,  0,  1 }, 	{ -18052,  11330,  18052,  2,  0,  1 },
	{ -17970,  11713,  17970,  2,  0,  1 }, 	{ -17886,  12094,  17886,  2,  0,  1 },
	{ -17799,  12474,  17799,  2,  0,  1 }, 	{ -17709,  12851,  17709,  2,  0,  1 },
	{ -17617,  13226,  17617,  2,  0,  1 }, 	{ -17522,  13600,  17522,  2,  0,  1 },
	{ -17424,  13971,  17424,  2,  0,  1 }, 	{ -17324,  14341,  17324,  2,  0,  1 },
	{ -17221,  14708,  17221,  2,  0,  1 }, 	{ -17115,  15073,  17115,  2,  0,  1 },
	{ -17007,  15435,  17007,  2,  0,  1 }, 	{ -16896,  15796,  16896,  2,  0,  1 },
	{ -16783,  16154,  16783,  2,  0,  1 }, 	{ -16668,  16509,  16668,  2,  0,  1 },
	{ -16706,  16393,  16706,  2,  1,  0 }, 	{ -16821,  16037,  16821,  2,  1,  0 },
	{ -16933,  15678,  16933,  2,  1,  0 }, 	{ -17043,  15317,  17043,  2,  1,  0 },
	{ -17150,  14953,  17150,  2,  1,  0 }, 	{ -17255,  14588,  17255,  2,  1,  0 },
	{ -17357,  14220,  17357,  2,  1,  0 }, 	{ -17456,  13850,  17456,  2,  1,  0 },
	{ -17553,  13478,  17553,  2,  1,  0 }, 	{ -17647,  13103,  17647,  2,  1,  0 },
	{ -17738,  12727,  17738,  2,  1,  0 }, 	{ -17827,  12349,  17827,  2,  1,  0 },
	{ -17913,  11969,  17913,  2,  1,  0 }, 	{ -17997,  11588,  17997,  2,  1,  0 },
	{ -18078,  11204,  18078,  2,  1,  0 }, 	{ -18156,  10819,  18156,  2,  1,  0 },
	{ -18231,  10433,  18231,  2,  1,  0 }, 	{ -18303,  10044,  18303,  2,  1,  0 },
	{ -18373,   9654,  18373,  2,  1,  0 }, 	{ -18440,   9263,  18440,  2,  1,  0 },
	{ -18505,   8870,  18505,  2,  1,  0 }, 	{ -18566,   8476,  18566,  2,  1,  0 },
	{ -18625,   8081,  18625,  2,  1,  0 }, 	{ -18680,   7685,  18680,  2,  1,  0 },
	{ -18733,   7287,  18733,  2,  1,  0 }, 	{ -18784,   6888,  18784,  2,  1,  0 },
	{ -18831,   6488,  18831,  2,  1,  0 }, 	{ -18876,   6088,  18876,  2,  1,  0 },
	{ -18917,   5686,  18917,  2,  1,  0 }, 	{ -18956,   5283,  18956,  2,  1,  0 },
	{ -18992,   4880,  18992,  2,  1,  0 }, 	{ -19025,   4476,  19025,  2,  1,  0 },
	{ -19056,   4071,  19056,  2,  1,  0 }, 	{ -19083,   3666,  19083,  2,  1,  0 },
	{ -19108,   3260,  19108,  2,  1,  0 }, 	{ -19129,   2853,  19129,  2,  1,  0 },
	{ -19148,   2447,  19148,  2,  1,  0 }, 	{ -19164,   2039,  19164,  2,  1,  0 },
	{ -19177,   1632,  19177,  2,  1,  0 }, 	{ -19187,   1224,  19187,  2,  1,  0 },
	{ -19194,    816,  19194,  2,  1,  0 }, 	{ -19199,    408,  19199,  2,  1,  0 },
	{ -19200,      0,  19200,  2,  1,  0 }, 	{ -19199,   -408,  19199,  2,  1,  0 },
	{ -19194,   -816,  19194,  2,  1,  0 }, 	{ -19187,  -1224,  19187,  2,  1,  0 },
	{ -19177,  -1632,  19177,  2,  1,  0 }, 	{ -19164,  -2039,  19164,  2,  1,  0 },
	{ -19148,  -2447,  19148,  2,  1,  0 }, 	{ -19129,  -2853,  19129,  2,  1,  0 },
	{ -19108,  -3260,  19108,  2,  1,  0 }, 	{ -19083,  -3666,  19083,  2,  1,  0 },
	{ -19056,  -4071,  19056,  2,  1,  0 }, 	{ -19025,  -4476,  19025,  2,  1,  0 },
	{ -18992,  -4880,  18992,  2,  1,  0 }, 	{ -18956,  -5283,  18956,  2,  1,  0 },
	{ -18917,  -5686,  18917,  2,  1,  0 }, 	{ -18876,  -6088,  18876,  2,  1,  0 },
	{ -18831,  -6488,  18831,  2,  1,  0 }, 	{ -18784,  -6888,  18784,  2,  1,  0 },
	{ -18733,  -7287,  18733,  2,  1,  0 }, 	{ -18680,  -7685,  18680,  2,  1,  0 },
	{ -18625,  -8081,  18625,  2,  1,  0 }, 	{ -18566,  -8476,  18566,  2,  1,  0 },
	{ -18505,  -8870,  18505,  2,  1,  0 }, 	{ -18440,  -9263,  18440,  2,  1,  0 },
	{ -18373,  -9654,  18373,  2,  1,  0 }, 	{ -18303, -10044,  18303,  2,  1,  0 },
	{ -18231, -10433,  18231,  2,  1,  0 }, 	{ -18156, -10819,  18156,  2,  1,  0 },
	{ -18078, -11204,  18078,  2,  1,  0 }, 	{ -17997, -11588,  17997,  2,  1,  0 },
	{ -17913, -11969,  17913,  2,  1,  0 }, 	{ -17827, -12349,  17827,  2,  1,  0 },
	{ -17738, -12727,  17738,  2,  1,  0 }, 	{ -17647, -13103,  17647,  2,  1,  0 },
	{ -17553, -13478,  17553,  2,  1,  0 }, 	{ -17456, -13850,  17456,  2,  1,  0 },
	{ -17357, -14220,  17357,  2,  1,  0 }, 	{ -17255, -14588,  17255,  2,  1,  0 },
	{ -17150, -14953,  17150,  2,  1,  0 }, 	{ -17043, -15317,  17043,  2,  1,  0 },
	{ -16933, -15678,  16933,  2,  1,  0 }, 	{ -16821, -16037,  16821,  2,  1,  0 },
	{ -16706, -16393,  16706,  2,  1,  0 }, 	{ -16668, -16509,  16668,  1,  2,  0 },
	{ -16783, -16154,  16783,  1,  2,  0 }, 	{ -16896, -15796,  16896,  1,  2,  0 },
	{ -17007, -15435,  17007,  1,  2,  0 }, 	{ -17115, -15073,  17115,  1,  2,  0 },
	{ -17221, -14708,  17221,  1,  2,  0 }, 	{ -17324, -14341,  17324,  1,  2,  0 },
	{ -17424, -13971,  17424,  1,  2,  0 }, 	{ -17522, -13600,  17522,  1,  2,  0 },
	{ -17617, -13226,  17617,  1,  2,  0 }, 	{ -17709, -12851,  17709,  1,  2,  0 },
	{ -17799, -12474,  17799,  1,  2,  0 }, 	{ -17886, -12094,  17886,  1,  2,  0 },
	{ -17970, -11713,  17970,  1,  2,  0 }, 	{ -18052, -11330,  18052,  1,  2,  0 },
	{ -18131, -10946,  18131,  1,  2,  0 }, 	{ -18207, -10560,  18207,  1,  2,  0 },
	{ -18281, -10172,  18281,  1,  2,  0 }, 	{ -18351,  -9783,  18351,  1,  2,  0 },
	{ -18419,  -9392,  18419,  1,  2,  0 }, 	{ -18484,  -9000,  18484,  1,  2,  0 },
	{ -18547,  -8606,  18547,  1,  2,  0 }, 	{ -18606,  -8211,  18606,  1,  2,  0 },
	{ -18663,  -7815,  18663,  1,  2,  0 }, 	{ -18717,  -7418,  18717,  1,  2,  0 },
	{ -18768,  -7020,  18768,  1,  2,  0 }, 	{ -18817,  -6620,  18817,  1,  2,  0 },
	{ -18862,  -6220,  18862,  1,  2,  0 }, 	{ -18905,  -5818,  18905,  1,  2,  0 },
	{ -18945,  -5416,  18945,  1,  2,  0 }, 	{ -18982,  -5013,  18982,  1,  2,  0 },
	{ -19016,  -4609,  19016,  1,  2,  0 }, 	{ -19047,  -4205,  19047,  1,  2,  0 },
	{ -19075,  -3800,  19075,  1,  2,  0 }, 	{ -19101,  -3394,  19101,  1,  2,  0 },
	{ -19123,  -2988,  19123,  1,  2,  0 }, 	{ -19143,  -2581,  19143,  1,  2,  0 },
	{ -19160,  -2174,  19160,  1,  2,  0 }, 	{ -19174,  -1766,  19174,  1,  2,  0 },
	{ -19185,  -1359,  19185,  1,  2,  0 }, 	{ -19193,   -951,  19193,  1,  2,  0 },
	{ -19199,   -543,  19199,  1,  2,  0 }, 	{ -19201,   -135,  19201,  1,  2,  0 },
	{ -19201,    273,  19201,  1,  2,  0 }, 	{ -19197,    681,  19197,  1,  2,  0 },
	{ -19191,   1089,  19191,  1,  2,  0 }, 	{ -19182,   1497,  19182,  1,  2,  0 },
	{ -19170,   1905,  19170,  1,  2,  0 }, 	{ -19155,   2312,  19155,  1,  2,  0 },
	{ -19137,   2719,  19137,  1,  2,  0 }, 	{ -19116,   3125,  19116,  1,  2,  0 },
	{ -19093,   3532,  19093,  1,  2,  0 }, 	{ -19066,   3937,  19066,  1,  2,  0 },
	{ -19037,   4342,  19037,  1,  2,  0 }, 	{ -19005,   4746,  19005,  1,  2,  0 },
	{ -18970,   5150,  18970,  1,  2,  0 }, 	{ -18932,   5553,  18932,  1,  2,  0 },
	{ -18891,   5955,  18891,  1,  2,  0 }, 	{ -18848,   6356,  18848,  1,  2,  0 },
	{ -18801,   6756,  18801,  1,  2,  0 }, 	{ -18752,   7155,  18752,  1,  2,  0 },
	{ -18700,   7553,  18700,  1,  2,  0 }, 	{ -18645,   7950,  18645,  1,  2,  0 },
	{ -18587,   8345,  18587,  1,  2,  0 }, 	{ -18527,   8740,  18527,  1,  2,  0 },
	{ -18463,   9133,  18463,  1,  2,  0 }, 	{ -18397,   9525,  18397,  1,  2,  0 },
	{ -18328,   9915,  18328,  1,  2,  0 }, 	{ -18257,  10304,  18257,  1,  2,  0 },
	{ -18182,  10691,  18182,  1,  2,  0 }, 	{ -18105,  11077,  18105,  1,  2,  0 },
	{ -18026,  11461,  18026,  1,  2,  0 }, 	{ -17943,  11843,  17943,  1,  2,  0 },
	{ -17858,  12223,  17858,  1,  2,  0 }, 	{ -17770,  12602,  17770,  1,  2,  0 },
	{ -17679,  12979,  17679,  1,  2,  0 }, 	{ -17586,  13353,  17586,  1,  2,  0 },
	{ -17490,  13726,  17490,  1,  2,  0 }, 	{ -17391,  14097,  17391,  1,  2,  0 },
	{ -17290,  14465,  17290,  1,  2,  0 }, 	{ -17187,  14832,  17187,  1,  2,  0 },
	{ -17080,  15196,  17080,  1,  2,  0 }, 	{ -16971,  15558,  16971,  1,  2,  0 },
	{ -16860,  15917,  16860,  1,  2,  0 }, 	{ -16746,  16274,  16746,  1,  2,  0 },
	{ -16629,  16629,  16629,  1,  2,  0 }, 	{ -16746,  16274,  16746,  1,  0,  2 },
	{ -16860,  15917,  16860,  1,  0,  2 }, 	{ -16971,  15558,  16971,  1,  0,  2 },
	{ -17080,  15196,  17080,  1,  0,  2 }, 	{ -17187,  14832,  17187,  1,  0,  2 },
	{ -17290,  14465,  17290,  1,  0,  2 }, 	{ -17391,  14097,  17391,  1,  0,  2 },
	{ -17490,  13726,  17490,  1,  0,  2 }, 	{ -17586,  13353,  17586,  1,  0,  2 },
	{ -17679,  12979,  17679,  1,  0,  2 }, 	{ -17770,  12602,  17770,  1,  0,  2 },
	{ -17858,  12223,  17858,  1,  0,  2 }, 	{ -17943,  11843,  17943,  1,  0,  2 },
	{ -18026,  11461,  18026,  1,  0,  2 }, 	{ -18105,  11077,  18105,  1,  0,  2 },
	{ -18182,  10691,  18182,  1,  0,  2 }, 	{ -18257,  10304,  18257,  1,  0,  2 },
	{ -18328,   9915,  18328,  1,  0,  2 }, 	{ -18397,   9525,  18397,  1,  0,  2 },
	{ -18463,   9133,  18463,  1,  0,  2 }, 	{ -18527,   8740,  18527,  1,  0,  2 },
	{ -18587,   8345,  18587,  1,  0,  2 }, 	{ -18645,   7950,  18645,  1,  0,  2 },
	{ -18700,   7553,  18700,  1,  0,  2 }, 	{ -18752,   7155,  18752,  1,  0,  2 },
	{ -18801,   6756,  18801,  1,  0,  2 }, 	{ -18848,   6356,  18848,  1,  0,  2 },
	{ -18891,   5955,  18891,  1,  0,  2 }, 	{ -18932,   5553,  18932,  1,  0,  2 },
	{ -18970,   5150,  18970,  1,  0,  2 }, 	{ -19005,   4746,  19005,  1,  0,  2 },
	{ -19037,   4342,  19037,  1,  0,  2 }, 	{ -19066,   3937,  19066,  1,  0,  2 },
	{ -19093,   3532,  19093,  1,  0,  2 }, 	{ -19116,   3125,  19116,  1,  0,  2 },
	{ -19137,   2719,  19137,  1,  0,  2 }, 	{ -19155,   2312,  19155,  1,  0,  2 },
	{ -19170,   1905,  19170,  1,  0,  2 }, 	{ -19182,   1497,  19182,  1,  0,  2 },
	{ -19191,   1089,  19191,  1,  0,  2 }, 	{ -19197,    681,  19197,  1,  0,  2 },
	{ -19201,    273,  19201,  1,  0,  2 }, 	{ -19201,   -135,  19201,  1,  0,  2 },
	{ -19199,   -543,  19199,  1,  0,  2 }, 	{ -19193,   -951,  19193,  1,  0,  2 },
	{ -19185,  -1359,  19185,  1,  0,  2 }, 	{ -19174,  -1766,  19174,  1,  0,  2 },
	{ -19160,  -2174,  19160,  1,  0,  2 }, 	{ -19143,  -2581,  19143,  1,  0,  2 },
	{ -19123,  -2988,  19123,  1,  0,  2 }, 	{ -19101,  -3394,  19101,  1,  0,  2 },
	{ -19075,  -3800,  19075,  1,  0,  2 }, 	{ -19047,  -4205,  19047,  1,  0,  2 },
	{ -19016,  -4609,  19016,  1,  0,  2 }, 	{ -18982,  -5013,  18982,  1,  0,  2 },
	{ -18945,  -5416,  18945,  1,  0,  2 }, 	{ -18905,  -5818,  18905,  1,  0,  2 },
	{ -18862,  -6220,  18862,  1,  0,  2 }, 	{ -18817,  -6620,  18817,  1,  0,  2 },
	{ -18768,  -7020,  18768,  1,  0,  2 }, 	{ -18717,  -7418,  18717,  1,  0,  2 },
	{ -18663,  -7815,  18663,  1,  0,  2 }, 	{ -18606,  -8211,  18606,  1,  0,  2 },
	{ -18547,  -8606,  18547,  1,  0,  2 }, 	{ -18484,  -9000,  18484,  1,  0,  2 },
	{ -18419,  -9392,  18419,  1,  0,  2 }, 	{ -18351,  -9783,  18351,  1,  0,  2 },
	{ -18281, -10172,  18281,  1,  0,  2 }, 	{ -18207, -10560,  18207,  1,  0,  2 },
	{ -18131, -10946,  18131,  1,  0,  2 }, 	{ -18052, -11330,  18052,  1,  0,  2 },
	{ -17970, -11713,  17970,  1,  0,  2 }, 	{ -17886, -12094,  17886,  1,  0,  2 },
	{ -17799, -12474,  17799,  1,  0,  2 }, 	{ -17709, -12851,  17709,  1,  0,  2 },
	{ -17617, -13226,  17617,  1,  0,  2 }, 	{ -17522, -13600,  17522,  1,  0,  2 },
	{ -17424, -13971,  17424,  1,  0,  2 }, 	{ -17324, -14341,  17324,  1,  0,  2 },
	{ -17221, -14708,  17221,  1,  0,  2 }, 	{ -17115, -15073,  17115,  1,  0,  2 },
	{ -17007, -15435,  17007,  1,  0,  2 }, 	{ -16896, -15796,  16896,  1,  0,  2 },
	{ -16783, -16154,  16783,  1,  0,  2 }, 	{ -16668, -16509,  16668,  1,  0,  2 },
	{ -16706, -16393,  16706,  0,  1,  2 }, 	{ -16821, -16037,  16821,  0,  1,  2 },
	{ -16933, -15678,  16933,  0,  1,  2 }, 	{ -17043, -15317,  17043,  0,  1,  2 },
	{ -17150, -14953,  17150,  0,  1,  2 }, 	{ -17255, -14588,  17255,  0,  1,  2 },
	{ -17357, -14220,  17357,  0,  1,  2 }, 	{ -17456, -13850,  17456,  0,  1,  2 },
	{ -17553, -13478,  17553,  0,  1,  2 }, 	{ -17647, -13103,  17647,  0,  1,  2 },
	{ -17738, -12727,  17738,  0,  1,  2 }, 	{ -17827, -12349,  17827,  0,  1,  2 },
	{ -17913, -11969,  17913,  0,  1,  2 }, 	{ -17997, -11588,  17997,  0,  1,  2 },
	{ -18078, -11204,  18078,  0,  1,  2 }, 	{ -18156, -10819,  18156,  0,  1,  2 },
	{ -18231, -10433,  18231,  0,  1,  2 }, 	{ -18303, -10044,  18303,  0,  1,  2 },
	{ -18373,  -9654,  18373,  0,  1,  2 }, 	{ -18440,  -9263,  18440,  0,  1,  2 },
	{ -18505,  -8870,  18505,  0,  1,  2 }, 	{ -18566,  -8476,  18566,  0,  1,  2 },
	{ -18625,  -8081,  18625,  0,  1,  2 }, 	{ -18680,  -7685,  18680,  0,  1,  2 },
	{ -18733,  -7287,  18733,  0,  1,  2 }, 	{ -18784,  -6888,  18784,  0,  1,  2 },
	{ -18831,  -6488,  18831,  0,  1,  2 }, 	{ -18876,  -6088,  18876,  0,  1,  2 },
	{ -18917,  -5686,  18917,  0,  1,  2 }, 	{ -18956,  -5283,  18956,  0,  1,  2 },
	{ -18992,  -4880,  18992,  0,  1,  2 }, 	{ -19025,  -4476,  19025,  0,  1,  2 },
	{ -19056,  -4071,  19056,  0,  1,  2 }, 	{ -19083,  -3666,  19083,  0,  1,  2 },
	{ -19108,  -3260,  19108,  0,  1,  2 }, 	{ -19129,  -2853,  19129,  0,  1,  2 },
	{ -19148,  -2447,  19148,  0,  1,  2 }, 	{ -19164,  -2039,  19164,  0,  1,  2 },
	{ -19177,  -1632,  19177,  0,  1,  2 }, 	{ -19187,  -1224,  19187,  0,  1,  2 },
	{ -19194,   -816,  19194,  0,  1,  2 }, 	{ -19199,   -408,  19199,  0,  1,  2 },
	{ -19200,      0,  19200,  0,  1,  2 }, 	{ -19199,    408,  19199,  0,  1,  2 },
	{ -19194,    816,  19194,  0,  1,  2 }, 	{ -19187,   1224,  19187,  0,  1,  2 },
	{ -19177,   1632,  19177,  0,  1,  2 }, 	{ -19164,   2039,  19164,  0,  1,  2 },
	{ -19148,   2447,  19148,  0,  1,  2 }, 	{ -19129,   2853,  19129,  0,  1,  2 },
	{ -19108,   3260,  19108,  0,  1,  2 }, 	{ -19083,   3666,  19083,  0,  1,  2 },
	{ -19056,   4071,  19056,  0,  1,  2 }, 	{ -19025,   4476,  19025,  0,  1,  2 },
	{ -18992,   4880,  18992,  0,  1,  2 }, 	{ -18956,   5283,  18956,  0,  1,  2 },
	{ -18917,   5686,  18917,  0,  1,  2 }, 	{ -18876,   6088,  18876,  0,  1,  2 },
	{ -18831,   6488,  18831,  0,  1,  2 }, 	{ -18784,   6888,  18784,  0,  1,  2 },
	{ -18733,   7287,  18733,  0,  1,  2 }, 	{ -18680,   7685,  18680,  0,  1,  2 },
	{ -18625,   8081,  18625,  0,  1,  2 }, 	{ -18566,   8476,  18566,  0,  1,  2 },
	{ -18505,   8870,  18505,  0,  1,  2 }, 	{ -18440,   9263,  18440,  0,  1,  2 },
	{ -18373,   9654,  18373,  0,  1,  2 }, 	{ -18303,  10044,  18303,  0,  1,  2 },
	{ -18231,  10433,  18231,  0,  1,  2 }, 	{ -18156,  10819,  18156,  0,  1,  2 },
	{ -18078,  11204,  18078,  0,  1,  2 }, 	{ -17997,  11588,  17997,  0,  1,  2 },
	{ -17913,  11969,  17913,  0,  1,  2 }, 	{ -17827,  12349,  17827,  0,  1,  2 },
	{ -17738,  12727,  17738,  0,  1,  2 }, 	{ -17647,  13103,  17647,  0,  1,  2 },
	{ -17553,  13478,  17553,  0,  1,  2 }, 	{ -17456,  13850,  17456,  0,  1,  2 },
	{ -17357,  14220,  17357,  0,  1,  2 }, 	{ -17255,  14588,  17255,  0,  1,  2 },
	{ -17150,  14953,  17150,  0,  1,  2 }, 	{ -17043,  15317,  17043,  0,  1,  2 },
	{ -16933,  15678,  16933,  0,  1,  2 }, 	{ -16821,  16037,  16821,  0,  1,  2 },
	{ -16706,  16393,  16706,  0,  1,  2 }, 	{ -16668,  16509,  16668,  0,  2,  1 },
	{ -16783,  16154,  16783,  0,  2,  1 }, 	{ -16896,  15796,  16896,  0,  2,  1 },
	{ -17007,  15435,  17007,  0,  2,  1 }, 	{ -17115,  15073,  17115,  0,  2,  1 },
	{ -17221,  14708,  17221,  0,  2,  1 }, 	{ -17324,  14341,  17324,  0,  2,  1 },
	{ -17424,  13971,  17424,  0,  2,  1 }, 	{ -17522,  13600,  17522,  0,  2,  1 },
	{ -17617,  13226,  17617,  0,  2,  1 }, 	{ -17709,  12851,  17709,  0,  2,  1 },
	{ -17799,  12474,  17799,  0,  2,  1 }, 	{ -17886,  12094,  17886,  0,  2,  1 },
	{ -17970,  11713,  17970,  0,  2,  1 }, 	{ -18052,  11330,  18052,  0,  2,  1 },
	{ -18131,  10946,  18131,  0,  2,  1 }, 	{ -18207,  10560,  18207,  0,  2,  1 },
	{ -18281,  10172,  18281,  0,  2,  1 }, 	{ -18351,   9783,  18351,  0,  2,  1 },
	{ -18419,   9392,  18419,  0,  2,  1 }, 	{ -18484,   9000,  18484,  0,  2,  1 },
	{ -18547,   8606,  18547,  0,  2,  1 }, 	{ -18606,   8211,  18606,  0,  2,  1 },
	{ -18663,   7815,  18663,  0,  2,  1 }, 	{ -18717,   7418,  18717,  0,  2,  1 },
	{ -18768,   7020,  18768,  0,  2,  1 }, 	{ -18817,   6620,  18817,  0,  2,  1 },
	{ -18862,   6220,  18862,  0,  2,  1 }, 	{ -18905,   5818,  18905,  0,  2,  1 },
	{ -18945,   5416,  18945,  0,  2,  1 }, 	{ -18982,   5013,  18982,  0,  2,  1 },
	{ -19016,   4609,  19016,  0,  2,  1 }, 	{ -19047,   4205,  19047,  0,  2,  1 },
	{ -19075,   3800,  19075,  0,  2,  1 }, 	{ -19101,   3394,  19101,  0,  2,  1 },
	{ -19123,   2988,  19123,  0,  2,  1 }, 	{ -19143,   2581,  19143,  0,  2,  1 },
	{ -19160,   2174,  19160,  0,  2,  1 }, 	{ -19174,   1766,  19174,  0,  2,  1 },
	{ -19185,   1359,  19185,  0,  2,  1 }, 	{ -19193,    951,  19193,  0,  2,  1 },
	{ -19199,    543,  19199,  0,  2,  1 }, 	{ -19201,    135,  19201,  0,  2,  1 },
	{ -19201,   -273,  19201,  0,  2,  1 }, 	{ -19197,   -681,  19197,  0,  2,  1 },
	{ -19191,  -1089,  19191,  0,  2,  1 }, 	{ -19182,  -1497,  19182,  0,  2,  1 },
	{ -19170,  -1905,  19170,  0,  2,  1 }, 	{ -19155,  -2312,  19155,  0,  2,  1 },
	{ -19137,  -2719,  19137,  0,  2,  1 }, 	{ -19116,  -3125,  19116,  0,  2,  1 },
	{ -19093,  -3532,  19093,  0,  2,  1 }, 	{ -19066,  -3937,  19066,  0,  2,  1 },
	{ -19037,  -4342,  19037,  0,  2,  1 }, 	{ -19005,  -4746,  19005,  0,  2,  1 },
	{ -18970,  -5150,  18970,  0,  2,  1 }, 	{ -18932,  -5553,  18932,  0,  2,  1 },
	{ -18891,  -5955,  18891,  0,  2,  1 }, 	{ -18848,  -6356,  18848,  0,  2,  1 },
	{ -18801,  -6756,  18801,  0,  2,  1 }, 	{ -18752,  -7155,  18752,  0,  2,  1 },
	{ -18700,  -7553,  18700,  0,  2,  1 }, 	{ -18645,  -7950,  18645,  0,  2,  1 },
	{ -18587,  -8345,  18587,  0,  2,  1 }, 	{ -18527,  -8740,  18527,  0,  2,  1 },
	{ -18463,  -9133,  18463,  0,  2,  1 }, 	{ -18397,  -9525,  18397,  0,  2,  1 },
	{ -18328,  -9915,  18328,  0,  2,  1 }, 	{ -18257, -10304,  18257,  0,  2,  1 },
	{ -18182, -10691,  18182,  0,  2,  1 }, 	{ -18105, -11077,  18105,  0,  2,  1 },
	{ -18026, -11461,  18026,  0,  2,  1 }, 	{ -17943, -11843,  17943,  0,  2,  1 },
	{ -17858, -12223,  17858,  0,  2,  1 }, 	{ -17770, -12602,  17770,  0,  2,  1 },
	{ -17679, -12979,  17679,  0,  2,  1 }, 	{ -17586, -13353,  17586,  0,  2,  1 },
	{ -17490, -13726,  17490,  0,  2,  1 }, 	{ -17391, -14097,  17391,  0,  2,  1 },
	{ -17290, -14465,  17290,  0,  2,  1 }, 	{ -17187, -14832,  17187,  0,  2,  1 },
	{ -17080, -15196,  17080,  0,  2,  1 }, 	{ -16971, -15558,  16971,  0,  2,  1 },
	{ -16860, -15917,  16860,  0,  2,  1 }, 	{ -16746, -16274,  16746,  0,  2,  1 },
};
