/*
 * prog_client.c
 *
 *  Created on: 03.04.2015
 *      Author: AndreR
 */

#include "prog_client.h"

#include "intercom_constants.h"
#include "util/intercom.h"
#include "util/flash.h"
#include "util/boot.h"
#include "util/crc.h"
#include "util/log.h"

ProgClient progClient;

static void sendNACK(uint16_t seq)
{
	BootloaderSeq bSeq = {0xFF, seq};

	IntercomHeader header;
	header.section = INTERCOM_SECTION_BOOTLOADER;
	header.command = BOOTLOADER_NACK;

	IntercomSend(&header, sizeof(BootloaderSeq), &bSeq);
}

static void sendACK(uint16_t seq)
{
	BootloaderSeq bSeq = {0xFF, seq};

	IntercomHeader header;
	header.section = INTERCOM_SECTION_BOOTLOADER;
	header.command = BOOTLOADER_ACK;

	IntercomSend(&header, sizeof(BootloaderSeq), &bSeq);
}

static void progClientCb(const IntercomHeader* pHeader, uint16_t dataLength, const uint8_t* pData, void* pUserData)
{
	(void)dataLength;
	(void)pUserData;

	switch(pHeader->command)
	{
		case BOOTLOADER_ERASE:
		{
			BootloaderSeq* pSeq = (BootloaderSeq*)pData;

			if(pSeq->index >= progClient.usedPrograms)
			{
				sendNACK(pSeq->seq);
				return;
			}

			ProgClientProgram* pProgram = &progClient.programs[pSeq->index];

			if(pProgram->programCodeErased)
			{
				sendACK(pSeq->seq);
				return;
			}

			if(FlashErase(pProgram->address, pProgram->address + pProgram->size))
			{
				sendNACK(pSeq->seq);
			}
			else
			{
				sendACK(pSeq->seq);
				pProgram->programCodeErased = 1;
			}
		}
		break;
		case BOOTLOADER_DATA:
		{
			BootloaderProgramData* pProgData = (BootloaderProgramData*)pData;
			pData += sizeof(BootloaderProgramData);
			dataLength -= sizeof(BootloaderProgramData);

			if((dataLength % sizeof(uint32_t)) != 0)
			{
				sendNACK(pProgData->seq);
				return;
			}

			if(pProgData->index >= progClient.usedPrograms)
			{
				sendNACK(pProgData->seq);
				return;
			}

			const ProgClientProgram* pProgram = &progClient.programs[pProgData->index];

			if(FlashProgram(pProgram->address+pProgData->offset, (uint32_t*)pData, dataLength/4))
				sendNACK(pProgData->seq);
			else
				sendACK(pProgData->seq);
		}
		break;
		case BOOTLOADER_CHECKSUM:
		{
			BootloaderChecksum* pReq = (BootloaderChecksum*)pData;

			if(pReq->index >= progClient.usedPrograms)
			{
				sendNACK(pReq->seq);
				return;
			}

			const ProgClientProgram* pProgram = &progClient.programs[pReq->index];

			if(CRC32VerifyChecksumAlt((uint8_t*)(pProgram->address+pReq->start), pReq->end-pReq->start, pReq->crc) == CRC32_MAGIC_NUMBER)
				sendACK(pReq->seq);
			else
				sendNACK(pReq->seq);
		}
		break;
		case BOOTLOADER_SELECT:
		{
			BootSetBootloaderSelected(1);
		}
		break;
	}
}

void ProgClientInit()
{
	IntercomClient client;
	client.pUserData = 0;
	client.section = INTERCOM_SECTION_BOOTLOADER;
	client.rxCallback = &progClientCb;

	IntercomAddClient(client);
}

int16_t ProgClientAddProgram(uint32_t address, uint32_t size)
{
	if(progClient.usedPrograms >= PROG_CLIENT_MAX_PROGRAMS)
	{
		LogError("Max programs reached");
		return -1;
	}

	ProgClientProgram* pProgram = &progClient.programs[progClient.usedPrograms];
	pProgram->address = address;
	pProgram->size = size;
	pProgram->programCodeErased = 0;

	++progClient.usedPrograms;

	return 0;
}
